import numpy as np
from typing import List
from hub.core.compression import (
    compress_bytes,
    compress_multiple,
    decompress_bytes,
    decompress_multiple,
)
from hub.core.fast_forwarding import ffw_chunk
from hub.core.meta.encode.shape import ShapeEncoder
from hub.core.serialize import bytes_to_text, check_sample_shape
from hub.core.partial_sample import PartialSample
from hub.core.tiling.sample_tiles import SampleTiles
from hub.util.casting import intelligent_cast
from hub.util.compression import get_compression_ratio
from hub.util.exceptions import EmptyTensorError
from .base_chunk import BaseChunk, InputSample
from hub.core.serialize import infer_chunk_num_bytes
import hub


class ChunkCompressedChunk(BaseChunk):
    def __init__(self, *args, **kwargs):
        super(ChunkCompressedChunk, self).__init__(*args, **kwargs)
        if self.is_byte_compression:
            self.decompressed_bytes = bytearray(
                decompress_bytes(self._data_bytes, self.compression)
            )
        else:
            shapes = [
                self.shapes_encoder[i] for i in range(self.shapes_encoder.num_samples)
            ]
            self.decompressed_samples = decompress_multiple(self._data_bytes, shapes)
        self._changed = False
        self._compression_ratio = 0.5

    def extend_if_has_space(self, incoming_samples: List[InputSample], update_tensor_meta: bool = True) -> float:  # type: ignore
        self.prepare_for_write()
        if self.is_byte_compression:
            return self.extend_if_has_space_byte_compression(
                incoming_samples, update_tensor_meta=update_tensor_meta
            )
        return self.extend_if_has_space_image_compression(
            incoming_samples, update_tensor_meta=update_tensor_meta
        )

    def extend_if_has_space_byte_compression(
        self,
        incoming_samples: List[InputSample],
        update_tensor_meta: bool = True,
    ):
        num_samples = 0

        for i, incoming_sample in enumerate(incoming_samples):
            serialized_sample, shape = self.serialize_sample(
                incoming_sample,
                chunk_compression=self.compression,
                store_uncompressed_tiles=True,
            )

            if shape is not None:
                self.num_dims = self.num_dims or len(shape)
                check_sample_shape(shape, self.num_dims)

            # for tiles we do not need to check
            if isinstance(serialized_sample, SampleTiles):
                incoming_samples[i] = serialized_sample  # type: ignore
                if self.is_empty:
                    self.write_tile(serialized_sample)
                    num_samples += 0.5  # type: ignore
                    tile = serialized_sample.yield_uncompressed_tile()
                    if tile is not None:
                        self.decompressed_bytes = tile.tobytes()
                    self._changed = True
                break
            sample_nbytes = len(serialized_sample)

            recompressed = False  # This flag helps avoid double concatenation
            if (
                len(self.decompressed_bytes) + sample_nbytes  # type: ignore
            ) * self._compression_ratio > self.min_chunk_size:

                decompressed_bytes = self.decompressed_bytes
                new_decompressed = self.decompressed_bytes + serialized_sample  # type: ignore

                compressed_bytes = compress_bytes(
                    new_decompressed, compression=self.compression
                )
                num_compressed_bytes = len(compressed_bytes)
                tiling_threshold = self.tiling_threshold
                if num_compressed_bytes > self.min_chunk_size and not (
                    not decompressed_bytes
                    and (
                        tiling_threshold < 0 or num_compressed_bytes > tiling_threshold
                    )
                ):
                    break
                recompressed = True
                self.decompressed_bytes = new_decompressed
                self._compression_ratio /= 2
                self._data_bytes = compressed_bytes
                self._changed = False
            if not recompressed:
                self.decompressed_bytes += serialized_sample  # type: ignore

            self._changed = True
            self.register_in_meta_and_headers(
                sample_nbytes, shape, update_tensor_meta=update_tensor_meta
            )
            num_samples += 1
        return num_samples

    def extend_if_has_space_image_compression(
        self,
        incoming_samples: List[InputSample],
        update_tensor_meta: bool = True,
    ):
        num_samples = 0
        num_decompressed_bytes = sum(
            x.nbytes for x in self.decompressed_samples  # type: ignore
        )

        for i, incoming_sample in enumerate(incoming_samples):
            incoming_sample, shape = self.process_sample_img_compr(incoming_sample)

            if shape is not None and self.is_empty_tensor and len(shape) != 3:
                self.change_dimensionality(shape)

            if isinstance(incoming_sample, SampleTiles):
                incoming_samples[i] = incoming_sample  # type: ignore
                if self.is_empty:
                    self.write_tile(incoming_sample)
                    num_samples += 0.5  # type: ignore
                    tile = incoming_sample.yield_uncompressed_tile()
                    if tile is not None:
                        self.decompressed_samples = [tile]
                    self._changed = True
                break
            if (
                num_decompressed_bytes + incoming_sample.nbytes  # type: ignore
            ) * self._compression_ratio > self.min_chunk_size:
                decompressed_samples = self.decompressed_samples
                new_samples = decompressed_samples + [incoming_sample]  # type: ignore

                compressed_bytes = compress_multiple(
                    new_samples,  # type: ignore
                    compression=self.compression,
                )
                num_compressed_bytes = len(compressed_bytes)
                tiling_threshold = self.tiling_threshold
                if num_compressed_bytes > self.min_chunk_size and not (
                    not decompressed_samples
                    and (
                        tiling_threshold < 0 or num_compressed_bytes > tiling_threshold
                    )
                ):
                    break
                self._compression_ratio /= 2
                self._data_bytes = compressed_bytes
                self._changed = False

            shape = incoming_sample.shape  # type: ignore
            shape = self.normalize_shape(shape)

            self.num_dims = self.num_dims or len(shape)
            check_sample_shape(shape, self.num_dims)
            self.decompressed_samples.append(incoming_sample)  # type: ignore

            self._changed = True
            # Byte positions are not relevant for chunk wise image compressions, so incoming_num_bytes=None.
            self.register_in_meta_and_headers(
                None, shape, update_tensor_meta=update_tensor_meta
            )
            num_samples += 1
        return num_samples

    def _get_partial_sample_tile(self, as_bytes=None):
        if self.decompressed_samples or self.decompressed_bytes:
            return None
        if as_bytes is None:
            as_bytes = self.is_byte_compression
        return super(ChunkCompressedChunk, self)._get_partial_sample_tile(
            as_bytes=as_bytes
        )

    def read_sample(
        self,
        local_index: int,
        cast: bool = True,
        copy: bool = False,
        decompress: bool = True,
        is_tile: bool = False,
    ):
        if not decompress:
            raise NotImplementedError(
                "`decompress=False` is not supported by chunk compressed chunks as it can cause recompression."
            )
        if self.is_empty_tensor:
            raise EmptyTensorError(
                "This tensor has only been populated with empty samples. "
                "Need to add at least one non-empty sample before retrieving data."
            )
        partial_sample_tile = self._get_partial_sample_tile(as_bytes=False)
        if partial_sample_tile is not None:
            return partial_sample_tile
        if self.is_image_compression:
            return self.decompressed_samples[local_index]  # type: ignore

        decompressed = memoryview(self.decompressed_bytes)  # type: ignore
        if not is_tile and self.is_fixed_shape:
            shape = tuple(self.tensor_meta.min_shape)
            sb, eb = self.get_byte_positions(local_index)
            decompressed = decompressed[sb:eb]
        else:
            shape = self.shapes_encoder[local_index]
            if not self.byte_positions_encoder.is_empty():
                sb, eb = self.byte_positions_encoder[local_index]
                decompressed = decompressed[sb:eb]
        if self.is_text_like:
            return bytes_to_text(decompressed, self.htype)
        ret = np.frombuffer(decompressed, dtype=self.dtype).reshape(shape)
        if copy and not ret.flags["WRITEABLE"]:
            ret = ret.copy()
        return ret

    def update_sample(self, local_index: int, new_sample: InputSample):
        self.prepare_for_write()
        if self.is_byte_compression:
            self.update_sample_byte_compression(local_index, new_sample)
        else:
            self.update_sample_img_compression(local_index, new_sample)

    def update_sample_byte_compression(self, local_index: int, new_sample: InputSample):
        serialized_sample, shape = self.serialize_sample(
            new_sample, chunk_compression=self.compression, break_into_tiles=False
        )
        self.check_shape_for_update(shape)
        partial_sample_tile = self._get_partial_sample_tile()
        if partial_sample_tile is not None:
            self.decompressed_bytes = partial_sample_tile
        decompressed_buffer = self.decompressed_bytes

        new_data_uncompressed = self.create_updated_data(
            local_index, decompressed_buffer, serialized_sample
        )
        self.decompressed_bytes = new_data_uncompressed
        self._changed = True
        new_nb = (
            None if self.byte_positions_encoder.is_empty() else len(serialized_sample)
        )
        self.update_in_meta_and_headers(local_index, new_nb, shape)

    def update_sample_img_compression(self, local_index: int, new_sample: InputSample):
        if new_sample is None:
            if self.tensor_meta.max_shape:
                new_sample = np.ones(
                    (0,) * len(self.tensor_meta.max_shape), dtype=self.dtype
                )
            else:
                # earlier sample was also None, do nothing
                return

        new_sample = intelligent_cast(new_sample, self.dtype, self.htype)
        shape = new_sample.shape
        shape = self.normalize_shape(shape)
        if self.is_empty_tensor and len(shape) != 3:
            self.change_dimensionality(shape)
        self.check_shape_for_update(shape)
        partial_sample_tile = self._get_partial_sample_tile()
        if partial_sample_tile is not None:
            self.decompressed_samples = [partial_sample_tile]
        decompressed_samples = self.decompressed_samples

        decompressed_samples[local_index] = new_sample  # type: ignore
        self._changed = True
        self.update_in_meta_and_headers(local_index, None, shape)  # type: ignore

        self.data_bytes = bytearray(  # type: ignore
            compress_multiple(decompressed_samples, self.compression)  # type: ignore
        )
        self.update_in_meta_and_headers(local_index, None, shape)

    def process_sample_img_compr(self, sample):
        if sample is None:
            if self.tensor_meta.max_shape:
                shape = (0,) * len(self.tensor_meta.max_shape)
            else:
                # we assume 3d, later we reset dimensions if the assumption was wrong
                shape = (0, 0, 0)
            return np.ones(shape, dtype=self.dtype), None
        if isinstance(sample, SampleTiles):
            return sample, sample.tile_shape
        elif isinstance(sample, PartialSample):
            return (
                SampleTiles(
                    compression=self.compression,
                    chunk_size=self.min_chunk_size,
                    htype=self.htype,
                    sample_shape=sample.sample_shape,
                    tile_shape=sample.tile_shape,
                    dtype=sample.dtype,
                ),
                sample.sample_shape,
            )
        if isinstance(sample, hub.core.tensor.Tensor):
            sample = sample.numpy()
        sample = intelligent_cast(sample, self.dtype, self.htype)
        shape = sample.shape
        shape = self.normalize_shape(shape)
        if not self.is_empty_tensor:
            self.num_dims = self.num_dims or len(shape)
            check_sample_shape(shape, self.num_dims)

        ratio = get_compression_ratio(self.compression)
        approx_compressed_size = sample.nbytes * ratio

        if (
            self.tiling_threshold >= 0
            and approx_compressed_size > self.tiling_threshold
        ):
            sample = SampleTiles(
                sample,
                self.compression,
                self.tiling_threshold,
                store_uncompressed_tiles=True,
            )

        return sample, shape

    def pop(self, index):
        self.prepare_for_write()
        if self.is_byte_compression:
            sb, eb = self.byte_positions_encoder[index]
            self.decompressed_bytes = (
                self.decompressed_bytes[:sb] + self.decompressed_bytes[eb:]
            )
            self._data_bytes = compress_bytes(self.decompressed_bytes, self.compression)
        else:
            self.decompressed_samples.pop(index)
            self._data_bytes = compress_multiple(
                self.decompressed_samples, self.compression
            )
        if not self.shapes_encoder.is_empty():
            self.shapes_encoder.pop(index)
        if not self.byte_positions_encoder.is_empty():
            self.byte_positions_encoder.pop(index)
        self._changed = True

    def pop_multiple(self, num_samples):
        if self.is_byte_compression:
            total_samples = self.num_samples
            self.decompressed_bytes = self.decompressed_bytes[
                : self.byte_positions_encoder[total_samples - num_samples][0]
            ]
            self._data_bytes = compress_bytes(self.decompressed_bytes, self.compression)
        else:
            for _ in range(num_samples):
                self.decompressed_samples.pop()
            self._data_bytes = compress_multiple(
                self.decompressed_samples, self.compression
            )

        for _ in range(num_samples):
            if not self.shapes_encoder.is_empty():
                self.shapes_encoder.pop()
            if not self.byte_positions_encoder.is_empty():
                self.byte_positions_encoder.pop()
        self._changed = False

    def _compress(self):
        if self.is_byte_compression:
            self._data_bytes = compress_bytes(self.decompressed_bytes, self.compression)
        else:
            self._data_bytes = compress_multiple(
                self.decompressed_samples, self.compression
            )

    @property
    def data_bytes(self):
        if self._changed:
            self._compress()
            self._changed = False
        return self._data_bytes

    @data_bytes.setter
    def data_bytes(self, value):
        self._data_bytes = value

    @property
    def num_uncompressed_bytes(self):
        if self.is_byte_compression:
            return len(self.decompressed_bytes)
        return sum(x.nbytes for x in self.decompressed_samples)

    @property
    def nbytes(self):
        """Calculates the number of bytes `tobytes` will be without having to call `tobytes`. Used by `LRUCache` to determine if this chunk can be cached."""
        return infer_chunk_num_bytes(
            self.version,
            self.shapes_encoder.array,
            self.byte_positions_encoder.array,
            len_data=min(self.num_uncompressed_bytes, self.max_chunk_size),
        )

    def prepare_for_write(self):
        ffw_chunk(self)
        self.is_dirty = True

    @property
    def is_empty_tensor(self):
        if self.is_byte_compression:
            return super().is_empty_tensor
        return self.tensor_meta.max_shape == [0, 0, 0]

    def change_dimensionality(self, shape):
        if len(shape) != 2:
            raise ValueError(
                f"Only amples with shape (H, W) and (H, W, C) are supported in chunks with image compression, got {shape} instead."
            )
        self.tensor_meta.max_shape = list(shape)
        self.tensor_meta.min_shape = list(shape)
        self.num_dims = len(shape)
        empty_shape = (0,) * self.num_dims
        self.tensor_meta.update_shape_interval(empty_shape)
        self.tensor_meta.is_dirty = True
        num_samples = self.shapes_encoder.num_samples
        self.shapes_encoder = ShapeEncoder()
        self.shapes_encoder.register_samples((0,) * len(shape), num_samples)
        if self.decompressed_samples:
            for i, arr in enumerate(self.decompressed_samples):
                self.decompressed_samples[i] = arr.reshape((0,) * self.num_dims)



import io
from logging import warning
import hub
from hub.util.exceptions import (
    SampleCompressionError,
    SampleDecompressionError,
    UnsupportedCompressionError,
    CorruptedSampleError,
)
from hub.util.point_cloud import LAS_HEADER_FILED_NAME_TO_PARSER
from hub.compression import (
    get_compression_type,
    BYTE_COMPRESSION,
    VIDEO_COMPRESSION,
    AUDIO_COMPRESSION,
    POINT_CLOUD_COMPRESSION,
)
from typing import Union, Tuple, Sequence, List, Optional, BinaryIO
import numpy as np
from pathlib import Path
from PIL import Image, UnidentifiedImageError  # type: ignore
from io import BytesIO

import mmap
import struct
import sys
import re
import numcodecs.lz4  # type: ignore
from numpy.core.fromnumeric import compress  # type: ignore
import math

try:
    import av  # type: ignore

    _PYAV_INSTALLED = True
except ImportError:
    _PYAV_INSTALLED = False

try:
    import lz4.frame  # type: ignore

    _LZ4_INSTALLED = True
except ImportError:
    _LZ4_INSTALLED = False

if sys.byteorder == "little":
    _NATIVE_INT32 = "<i4"
    _NATIVE_FLOAT32 = "<f4"
else:
    _NATIVE_INT32 = ">i4"
    _NATIVE_FLOAT32 = ">f4"

DIMS_RE = re.compile(rb" ([0-9]+)x([0-9]+)")
FPS_RE = re.compile(rb" ([0-9]+) fps,")
DURATION_RE = re.compile(rb"Duration: ([0-9:.]+),")
INFO_RE = re.compile(rb"([a-z]+)=([0-9./]+)")

_JPEG_SOFS = [
    b"\xff\xc0",
    b"\xff\xc2",
    b"\xff\xc1",
    b"\xff\xc3",
    b"\xff\xc5",
    b"\xff\xc6",
    b"\xff\xc7",
    b"\xff\xc9",
    b"\xff\xca",
    b"\xff\xcb",
    b"\xff\xcd",
    b"\xff\xce",
    b"\xff\xcf",
    b"\xff\xde",
    # Skip:
    b"\xff\xcc",
    b"\xff\xdc",
    b"\xff\xdd",
    b"\xff\xdf",
    # App: (0xFFE0 - 0xFFEF):
    *map(lambda x: x.to_bytes(2, "big"), range(0xFFE0, 0xFFF0)),
    # DQT:
    b"\xff\xdb",
    # COM:
    b"\xff\xfe",
    # Start of scan
    b"\xff\xda",
]

_JPEG_SKIP_MARKERS = set(_JPEG_SOFS[14:])
_JPEG_SOFS_RE = re.compile(b"|".join(_JPEG_SOFS))
_STRUCT_HHB = struct.Struct(">HHB")
_STRUCT_II = struct.Struct(">ii")


def to_image(array: np.ndarray) -> Image:
    shape = array.shape
    if len(shape) == 3 and shape[0] != 1 and shape[2] == 1:
        # convert (X,Y,1) grayscale to (X,Y) for pillow compatibility
        return Image.fromarray(array.squeeze(axis=2))

    return Image.fromarray(array)


def _compress_apng(array: np.ndarray) -> bytes:
    if array.ndim == 3:
        # Binary APNG
        frames = list(
            map(Image.fromarray, (array[:, :, i] for i in range(array.shape[2])))
        )
    elif array.ndim == 4 and array.shape[3] <= 4:
        # RGB(A) APNG
        frames = list(map(Image.fromarray, array))
    else:
        raise SampleCompressionError(array.shape, "apng", "Unexpected shape.")
    out = BytesIO()
    frames[0].save(out, "png", save_all=True, append_images=frames[1:])
    out.seek(0)
    ret = out.read()
    out.close()
    return ret


def _decompress_apng(buffer: Union[bytes, memoryview]) -> np.ndarray:
    img = Image.open(BytesIO(buffer))
    frame0 = np.array(img)
    if frame0.ndim == 2:
        ret = np.zeros(frame0.shape + (img.n_frames,), dtype=frame0.dtype)
        ret[:, :, 0] = frame0
        for i in range(1, img.n_frames):
            img.seek(i)
            ret[:, :, i] = np.array(img)
    else:
        ret = np.zeros((img.n_frames,) + frame0.shape, dtype=frame0.dtype)
        ret[0] = frame0
        for i in range(1, img.n_frames):
            img.seek(i)
            ret[i] = np.array(img)
    return ret


def compress_bytes(
    buffer: Union[bytes, memoryview], compression: Optional[str]
) -> bytes:
    if not buffer:
        return b""
    if compression == "lz4":
        if not buffer:
            return b""
        return numcodecs.lz4.compress(buffer)
    else:
        raise SampleCompressionError(
            (len(buffer),), compression, f"Not a byte compression: {compression}"
        )


def decompress_bytes(
    buffer: Union[bytes, memoryview], compression: Optional[str]
) -> bytes:
    if not buffer:
        return b""
    if compression == "lz4":
        # weird edge case of lz4 + empty string
        if buffer == b"\x00\x00\x00\x00\x00":
            return b""
        if (
            buffer[:4] == b'\x04"M\x18'
        ):  # python-lz4 magic number (backward compatiblity)
            if not _LZ4_INSTALLED:
                raise ModuleNotFoundError(
                    "Module lz4 not found. Install using `pip install lz4`."
                )
            return lz4.frame.decompress(buffer)
        return numcodecs.lz4.decompress(buffer)
    else:
        raise SampleDecompressionError()


def compress_array(array: np.ndarray, compression: Optional[str]) -> bytes:
    """Compress some numpy array using `compression`. All meta information will be contained in the returned buffer.

    Note:
        `decompress_array` may be used to decompress from the returned bytes back into the `array`.

    Args:
        array (np.ndarray): Array to be compressed.
        compression (str, optional): `array` will be compressed with this compression into bytes. Right now only arrays compatible with `PIL` will be compressed.

    Raises:
        UnsupportedCompressionError: If `compression` is unsupported. See `hub.compressions`.
        SampleCompressionError: If there was a problem compressing `array`.
        NotImplementedError: If compression is not supported.

    Returns:
        bytes: Compressed `array` represented as bytes.
    """

    # empty sample shouldn't be compressed

    if 0 in array.shape:
        return bytes()

    if compression not in hub.compressions:
        raise UnsupportedCompressionError(compression)

    if compression is None:
        return array.tobytes()

    compr_type = get_compression_type(compression)

    if compr_type == BYTE_COMPRESSION:
        return compress_bytes(array.tobytes(), compression)
    elif compr_type == AUDIO_COMPRESSION:
        raise NotImplementedError(
            "In order to store audio data, you should use `hub.read(path_to_file)`. Compressing raw data is not yet supported."
        )
    elif compr_type == VIDEO_COMPRESSION:
        raise NotImplementedError(
            "In order to store video data, you should use `hub.read(path_to_file)`. Compressing raw data is not yet supported."
        )
    elif compr_type == POINT_CLOUD_COMPRESSION:
        raise NotImplementedError(
            "In order to store point cloud data, you should use `hub.read(path_to_file)`. Compressing raw data is not yet supported."
        )
    if compression == "apng":
        return _compress_apng(array)
    try:
        img = to_image(array)
        out = BytesIO()
        out._close = out.close  # type: ignore
        out.close = (  # type: ignore
            lambda: None
        )  # sgi save handler will try to close the stream (see https://github.com/python-pillow/Pillow/pull/5645)
        kwargs = {"sizes": [img.size]} if compression == "ico" else {}
        img.save(out, compression, **kwargs)
        out.seek(0)
        compressed_bytes = out.read()
        out._close()  # type: ignore
        return compressed_bytes
    except (TypeError, OSError) as e:
        raise SampleCompressionError(array.shape, compression, str(e))


def decompress_array(
    buffer: Union[bytes, memoryview, str],
    shape: Optional[Tuple[int, ...]] = None,
    dtype: Optional[str] = None,
    compression: Optional[str] = None,
    start_idx: Optional[int] = None,
    end_idx: Optional[int] = None,
    step: Optional[int] = None,
    reverse: bool = False,
) -> np.ndarray:
    """Decompress some buffer into a numpy array. It is expected that all meta information is
    stored inside `buffer`.

    Note:
        `compress_array` may be used to get the `buffer` input.

    Args:
        buffer (bytes, memoryview, str): Buffer or file to be decompressed. It is assumed all meta information required to
            decompress is contained within `buffer`, except for byte compressions
        shape (Tuple[int], Optional): Desired shape of decompressed object. Reshape will attempt to match this shape before returning.
        dtype (str, Optional): Applicable only for byte compressions. Expected dtype of decompressed array.
        compression (str, Optional): Applicable only for byte compressions. Compression used to compression the given buffer.
        start_idx: (int, Optional): Applicable only for video compressions. Index of first frame.
        end_idx: (int, Optional): Applicable only for video compressions. Index of last frame (exclusive).
        step: (int, Optional): Applicable only for video compressions. Step size for seeking.
        reverse (bool): Applicable only for video compressions. Reverses output numpy array if set to True.

    Raises:
        SampleDecompressionError: If decompression fails.
        ValueError: If dtype and shape are not specified for byte compression.

    Returns:
        np.ndarray: Array from the decompressed buffer.
    """
    compr_type = get_compression_type(compression)
    if compr_type == BYTE_COMPRESSION:
        if dtype is None or shape is None:
            raise ValueError("dtype and shape must be specified for byte compressions.")
        try:
            decompressed_bytes = decompress_bytes(buffer, compression)  # type: ignore
            return np.frombuffer(decompressed_bytes, dtype=dtype).reshape(shape)
        except Exception:
            raise SampleDecompressionError()
    elif compr_type == AUDIO_COMPRESSION:
        return _decompress_audio(buffer)
    elif compr_type == VIDEO_COMPRESSION:
        return _decompress_video(buffer, start_idx, end_idx, step, reverse)  # type: ignore
    elif compr_type == POINT_CLOUD_COMPRESSION:
        return _decompress_full_point_cloud(buffer)

    if compression == "apng":
        return _decompress_apng(buffer)  # type: ignore
    if compression == "dcm":
        return _decompress_dicom(buffer)  # type: ignore
    if compression is None and isinstance(buffer, memoryview) and shape is not None:
        assert buffer is not None
        assert shape is not None
        return np.frombuffer(buffer=buffer, dtype=dtype).reshape(shape)

    try:
        if shape is not None and 0 in shape:
            return np.zeros(shape, dtype=dtype)
        if not isinstance(buffer, str):
            buffer = BytesIO(buffer)  # type: ignore
        img = Image.open(buffer)  # type: ignore
        arr = np.array(img)
        if shape is not None:
            arr = arr.reshape(shape)
        return arr
    except Exception:
        raise SampleDecompressionError()


def _get_bounding_shape(shapes: Sequence[Tuple[int, ...]]) -> Tuple[int, int, int]:
    """Gets the shape of a bounding box that can contain the given the shapes tiled horizontally."""
    if len(shapes) == 0:
        return (0, 0, 0)
    channels_shape = None
    for shape in shapes:
        if shape != (0, 0, 0):
            channels_shape = shape[2:]
            break
    if channels_shape is None:
        channels_shape = (0,)
    for shape in shapes:
        if shape != (0, 0, 0) and shape[2:] != channels_shape:
            raise ValueError(
                "The data can't be compressed as the number of channels doesn't match."
            )
    return (max(s[0] for s in shapes), sum(s[1] for s in shapes)) + channels_shape  # type: ignore


def compress_multiple(
    arrays: Sequence[np.ndarray], compression: Optional[str]
) -> bytes:
    """Compress multiple arrays of different shapes into a single buffer. Used for chunk wise compression.
    The arrays are tiled horizontally and padded with zeros to fit in a bounding box, which is then compressed."""
    if len(arrays) == 0:
        return b""
    dtype = None
    for arr in arrays:
        if arr.size:
            if dtype is None:
                dtype = arr.dtype
            elif arr.dtype != dtype:
                raise SampleCompressionError(
                    arr.shape,
                    compression,
                    message="All arrays expected to have same dtype.",
                )

    compr_type = get_compression_type(compression)
    if compr_type == BYTE_COMPRESSION:
        return compress_bytes(
            b"".join(arr.tobytes() for arr in arrays), compression
        )  # Note: shape and dtype info not included
    elif compr_type == AUDIO_COMPRESSION:
        raise NotImplementedError("compress_multiple does not support audio samples.")
    elif compr_type == VIDEO_COMPRESSION:
        raise NotImplementedError("compress_multiple does not support video samples.")
    elif compr_type == POINT_CLOUD_COMPRESSION:
        raise NotImplementedError(
            "compress_multiple does not support point cloud samples."
        )
    elif compression == "apng":
        raise NotImplementedError("compress_multiple does not support apng samples.")
    canvas = np.zeros(_get_bounding_shape([arr.shape for arr in arrays]), dtype=dtype)
    next_x = 0
    for arr in arrays:
        if arr.shape == (0, 0, 0):
            continue
        canvas[: arr.shape[0], next_x : next_x + arr.shape[1]] = arr
        next_x += arr.shape[1]
    return compress_array(canvas, compression=compression)


def decompress_multiple(
    buffer: Union[bytes, memoryview],
    shapes: Sequence[Tuple[int, ...]],
    dtype: Optional[Union[np.dtype, str]] = None,
    compression: Optional[str] = None,
) -> List[np.ndarray]:
    """Unpack a compressed buffer into multiple arrays."""
    if not buffer:
        return []
    if compression and get_compression_type(compression) == "byte":
        decompressed_buffer = memoryview(decompress_bytes(buffer, compression))
        arrays = []
        itemsize = np.dtype(dtype).itemsize
        for shape in shapes:
            nbytes = int(np.prod(shape) * itemsize)
            arrays.append(
                np.frombuffer(decompressed_buffer[:nbytes], dtype=dtype).reshape(shape)
            )
            decompressed_buffer = decompressed_buffer[nbytes:]
        return arrays
    canvas = decompress_array(buffer)
    arrays = []
    next_x = 0
    for shape in shapes:
        if shape == (0, 0, 0):
            arrays.append(np.zeros(shape, dtype=canvas.dtype))
        else:
            arrays.append(canvas[: shape[0], next_x : next_x + shape[1]])
            next_x += shape[1]
    return arrays


def verify_compressed_file(
    file: Union[str, BinaryIO, bytes, memoryview], compression: str
) -> Union[Tuple[Tuple[int, ...], str], None]:
    """Verify the contents of an image file
    Args:
        file (Union[str, BinaryIO, bytes, memoryview]): Path to the file or file like object or contents of the file
        compression (str): Expected compression of the image file
    """
    if isinstance(file, str):
        file = open(file, "rb")
        close = True
    elif hasattr(file, "read"):
        close = False
        file.seek(0)  # type: ignore
    else:
        close = False
    try:
        if compression == "png":
            return _verify_png(file)
        elif compression == "jpeg":
            return _verify_jpeg(file), "|u1"
        elif get_compression_type(compression) == AUDIO_COMPRESSION:
            return _read_audio_shape(file, compression), "<f4"  # type: ignore
        elif compression in ("mp4", "mkv", "avi"):
            if isinstance(file, (bytes, memoryview, str)):
                return _read_video_shape(file), "|u1"  # type: ignore
        elif compression == "dcm":
            return _read_dicom_shape_and_dtype(file)
        elif compression == "las":
            return _read_point_cloud_shape_and_dtype(file)
        else:
            return _fast_decompress(file)
    except Exception as e:
        raise CorruptedSampleError(compression)
    finally:
        if close:
            file.close()  # type: ignore

    return None


def get_compression(header=None, path=None):
    if path:
        # These formats are recognized by file extension for now
        file_formats = [
            ".mp3",
            ".flac",
            ".wav",
            ".mp4",
            ".mkv",
            ".avi",
            ".dcm",
            ".las",
        ]
        path = str(path).lower()
        for fmt in file_formats:
            if path.endswith(fmt):
                return fmt[1:]
    if header:
        if not Image.OPEN:
            Image.init()
        for fmt in Image.OPEN:
            accept = Image.OPEN[fmt][1]
            if accept and accept(header):
                return fmt.lower()
        raise SampleDecompressionError()


def _verify_png(buf):
    if not hasattr(buf, "read"):
        buf = BytesIO(buf)
    img = Image.open(buf)
    img.verify()
    return Image._conv_type_shape(img)


def _verify_jpeg(f):
    if hasattr(f, "read"):
        return _verify_jpeg_file(f)
    return _verify_jpeg_buffer(f)


def _verify_jpeg_buffer(buf: bytes):
    # Start of Image
    mview = memoryview(buf)
    assert buf.startswith(b"\xff\xd8")
    # Look for Start of Frame
    sof_idx = -1
    offset = 0
    while True:
        match = _re_find_first(_JPEG_SOFS_RE, mview[offset:])
        if match is None:
            break
        idx = match.start(0) + offset
        marker = buf[idx : idx + 2]
        if marker == _JPEG_SOFS[-1]:
            break
        offset = idx + int.from_bytes(buf[idx + 2 : idx + 4], "big") + 2
        if marker not in _JPEG_SKIP_MARKERS:
            sof_idx = idx
    if sof_idx == -1:
        raise Exception()

    length = int.from_bytes(mview[sof_idx + 2 : sof_idx + 4], "big")
    assert mview[sof_idx + length + 2 : sof_idx + length + 4] in [
        b"\xff\xc4",
        b"\xff\xdb",
        b"\xff\xdd",
        b"\xff\xda",
    ]  # DHT, DQT, DRI, SOS
    shape = _STRUCT_HHB.unpack(mview[sof_idx + 5 : sof_idx + 10])
    assert buf.find(b"\xff\xd9") != -1
    if shape[-1] in (1, None):
        shape = shape[:-1]
    return shape


def _verify_jpeg_file(f):
    # See: https://dev.exiv2.org/projects/exiv2/wiki/The_Metadata_in_JPEG_files#2-The-metadata-structure-in-JPEG
    mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
    mv = memoryview(mm)
    try:
        soi = f.read(2)
        # Start of Image
        assert soi == b"\xff\xd8"

        # Look for Start of Frame
        sof_idx = -1
        offset = 0
        while True:
            view = mv[offset:]
            match = _re_find_first(_JPEG_SOFS_RE, view)
            view.release()
            if match is None:
                break
            idx = match.start(0) + offset
            marker = mm[idx : idx + 2]
            if marker == _JPEG_SOFS[-1]:
                break
            f.seek(idx + 2)
            offset = idx + int.from_bytes(f.read(2), "big") + 2
            if marker not in _JPEG_SKIP_MARKERS:
                sof_idx = idx
        if sof_idx == -1:
            raise Exception()  # Caught by verify_compressed_file()

        f.seek(sof_idx + 2)
        length = int.from_bytes(f.read(2), "big")
        f.seek(length - 2, 1)
        definition_start = f.read(2)
        assert definition_start in [
            b"\xff\xc4",
            b"\xff\xdb",
            b"\xff\xdd",
            b"\xff\xda",
        ]  # DHT, DQT, DRI, SOS
        f.seek(sof_idx + 5)
        shape = _STRUCT_HHB.unpack(f.read(5))
        # TODO this check is too slow
        assert mm.find(b"\xff\xd9") != -1  # End of Image
        if shape[-1] in (1, None):
            shape = shape[:-1]
        return shape
    finally:
        mv.release()
        mm.close()


def _fast_decompress(buf):
    if not hasattr(buf, "read"):
        buf = BytesIO(buf)
    img = Image.open(buf)
    img.load()
    if img.mode == 1:
        args = ("L",)
    else:
        args = (img.mode,)
    enc = Image._getencoder(img.mode, "raw", args)
    enc.setimage(img.im)
    bufsize = max(65536, img.size[0] * 4)
    while True:
        status, err_code, buf = enc.encode(
            bufsize
        )  # See https://github.com/python-pillow/Pillow/blob/master/src/encode.c#L144
        if err_code:
            break
    if err_code < 0:
        raise Exception()  # caught by verify_compressed_file()
    return Image._conv_type_shape(img)


def read_meta_from_compressed_file(
    file, compression: Optional[str] = None
) -> Tuple[str, Tuple[int], str]:
    """Reads shape, dtype and format without decompressing or verifying the sample."""
    if isinstance(file, (str, Path)):
        f = open(file, "rb")
        isfile = True
        close = True
    elif hasattr(file, "read"):
        f = file
        close = False
        isfile = True
        f.seek(0)
    else:
        isfile = False
        f = file
        close = False
    try:
        if compression is None:
            path = file if isinstance(file, str) else None
            if hasattr(f, "read"):
                compression = get_compression(f.read(32), path)
                f.seek(0)
            else:
                compression = get_compression(f[:32], path)  # type: ignore
        if compression == "jpeg":
            try:
                shape, typestr = _read_jpeg_shape(f), "|u1"
            except Exception:
                raise CorruptedSampleError("jpeg")
        elif compression == "png":
            try:
                shape, typestr = _read_png_shape_and_dtype(f)
            except Exception:
                raise CorruptedSampleError("png")
        elif compression == "dcm":
            shape, typestr = _read_dicom_shape_and_dtype(f)
        elif get_compression_type(compression) == AUDIO_COMPRESSION:
            try:
                shape, typestr = _read_audio_shape(file), "<f4"
            except Exception as e:
                raise CorruptedSampleError(compression)
        elif compression in ("mp4", "mkv", "avi"):
            try:
                shape, typestr = _read_video_shape(file), "|u1"  # type: ignore
            except Exception as e:
                raise CorruptedSampleError(compression)
        elif compression == "las":
            try:
                shape, typestr = _read_point_cloud_shape_and_dtype(file)
            except Exception as e:
                raise CorruptedSampleError(compression) from e
        else:
            img = Image.open(f) if isfile else Image.open(BytesIO(f))  # type: ignore
            shape, typestr = Image._conv_type_shape(img)
            compression = img.format.lower()
        return compression, shape, typestr  # type: ignore
    finally:
        if close:
            f.close()


def _read_jpeg_shape(f: Union[bytes, BinaryIO]) -> Tuple[int, ...]:
    if hasattr(f, "read"):
        return _read_jpeg_shape_from_file(f)
    return _read_jpeg_shape_from_buffer(f)  # type: ignore


def _re_find_first(pattern, string):
    for match in re.finditer(pattern, string):
        return match


def _read_jpeg_shape_from_file(f) -> Tuple[int, ...]:
    """Reads shape of a jpeg image from file without loading the whole image in memory"""
    mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_COPY)
    mv = memoryview(mm)
    try:
        # Look for Start of Frame
        sof_idx = -1
        offset = 0
        while True:
            view = mv[offset:]
            match = _re_find_first(_JPEG_SOFS_RE, view)
            view.release()
            if match is None:
                break
            idx = match.start(0) + offset
            marker = mm[idx : idx + 2]
            if marker == _JPEG_SOFS[-1]:
                break
            f.seek(idx + 2)
            offset = idx + int.from_bytes(f.read(2), "big") + 2
            if marker not in _JPEG_SKIP_MARKERS:
                sof_idx = idx
        if sof_idx == -1:
            raise Exception()
        f.seek(sof_idx + 5)
        shape = _STRUCT_HHB.unpack(f.read(5))  # type: ignore
        if shape[-1] in (1, None):
            shape = shape[:-1]
        return shape
    finally:
        mv.release()
        mm.close()


def _read_jpeg_shape_from_buffer(buf: bytes) -> Tuple[int, ...]:
    """Gets shape of a jpeg file from its contents"""
    # Look for Start of Frame
    mv = memoryview(buf)
    sof_idx = -1
    offset = 0
    while True:
        match = _re_find_first(_JPEG_SOFS_RE, mv[offset:])
        if match is None:
            break
        idx = match.start(0) + offset
        marker = buf[idx : idx + 2]
        if marker == _JPEG_SOFS[-1]:
            break
        offset = idx + int.from_bytes(buf[idx + 2 : idx + 4], "big") + 2
        if marker not in _JPEG_SKIP_MARKERS:
            sof_idx = idx
    if sof_idx == -1:
        raise Exception()
    shape = _STRUCT_HHB.unpack(memoryview(buf)[sof_idx + 5 : sof_idx + 10])  # type: ignore
    if shape[-1] in (1, None):
        shape = shape[:-1]
    return shape


def _read_dicom_shape_and_dtype(
    f: Union[bytes, BinaryIO]
) -> Tuple[Tuple[int, ...], str]:
    try:
        from pydicom import dcmread
        from pydicom.pixel_data_handlers.util import pixel_dtype
    except ImportError:
        raise ModuleNotFoundError(
            "Pydicom not found. Install using `pip install pydicom`"
        )
    if not hasattr(f, "read"):
        f = BytesIO(f)  # type: ignore
    dcm = dcmread(f)
    nchannels = dcm[0x0028, 0x0002].value
    shape = (dcm.Rows, dcm.Columns, nchannels)
    isfloat = "FloatPixelData" in dcm or "DoubleFloatPixelData" in dcm
    dtype = pixel_dtype(dcm, isfloat).str
    return shape, dtype


def _decompress_dicom(f: Union[str, bytes, BinaryIO]):
    if isinstance(f, (bytes, memoryview, bytearray)):
        f = BytesIO(f)
    try:
        from pydicom import dcmread
    except ImportError:
        raise ModuleNotFoundError(
            "Pydicom not found. Install using `pip install pydicom`"
        )
    arr = dcmread(f).pixel_array
    if arr.ndim == 2:
        return np.expand_dims(arr, -1)
    return arr


def _read_png_shape_and_dtype(f: Union[bytes, BinaryIO]) -> Tuple[Tuple[int, ...], str]:
    """Reads shape and dtype of a png file from a file like object or file contents.
    If a file like object is provided, all of its contents are NOT loaded into memory."""
    if not hasattr(f, "read"):
        f = BytesIO(f)  # type: ignore
    f.seek(16)  # type: ignore
    size = _STRUCT_II.unpack(f.read(8))[::-1]  # type: ignore
    bits, colors = f.read(2)  # type: ignore

    # Get the number of channels and dtype based on bits and colors:
    if colors == 0:
        if bits == 1:
            typstr = "|b1"
        elif bits == 16:
            typstr = _NATIVE_INT32
        else:
            typstr = "|u1"
        nlayers = None
    else:
        typstr = "|u1"
        if colors == 2:
            nlayers = 3
        elif colors == 3:
            nlayers = None
        elif colors == 4:
            if bits == 8:
                nlayers = 2
            else:
                nlayers = 4
        else:
            nlayers = 4
    shape = size if nlayers is None else size + (nlayers,)
    return shape, typstr  # type: ignore


def _frame_to_stamp(nframe, stream):
    """Convert frame number to timestamp based on fps of video stream."""
    fps = stream.guessed_rate.numerator / stream.guessed_rate.denominator
    seek_target = nframe / fps
    stamp = math.floor(
        seek_target * (stream.time_base.denominator / stream.time_base.numerator)
    )
    return stamp


def _open_video(file: Union[str, bytes, memoryview]):
    if not _PYAV_INSTALLED:
        raise ModuleNotFoundError(
            "PyAV is not installed. Run `pip install hub[video]`."
        )
    if isinstance(file, str):
        container = av.open(
            file, options={"protocol_whitelist": "file,http,https,tcp,tls,subfile"}
        )
    else:
        container = av.open(BytesIO(file))

    vstreams = container.streams.video

    if len(vstreams) == 0:
        raise IndexError("No video streams available!")

    vstream = vstreams[0]

    return container, vstream


def _read_metadata_from_vstream(container, vstream):
    duration = vstream.duration
    if duration is None:
        duration = container.duration
        time_base = 1 / av.time_base
    else:
        time_base = vstream.time_base.numerator / vstream.time_base.denominator
    fps = vstream.guessed_rate.numerator / vstream.guessed_rate.denominator

    nframes = vstream.frames
    if nframes == 0:
        nframes = math.floor(fps * duration * time_base)

    height = vstream.codec_context.height
    width = vstream.codec_context.width
    shape = (nframes, height, width, 3)

    return shape, duration, fps, time_base


def _read_video_shape(
    file: Union[str, bytes, memoryview],
):
    container, vstream = _open_video(file)
    shape = _read_metadata_from_vstream(container, vstream)[0]
    return shape


def _decompress_video(
    file: Union[str, bytes],
    start: int,
    stop: int,
    step: int,
    reverse: bool,
):
    container, vstream = _open_video(file)
    nframes, height, width, _ = _read_metadata_from_vstream(container, vstream)[0]

    if start is None:
        start = 0

    if stop is None:
        stop = nframes

    if step is None:
        step = 1

    nframes = math.ceil((stop - start) / step)

    video = np.zeros((nframes, height, width, 3), dtype=np.uint8)

    seek_target = _frame_to_stamp(start, vstream)
    step_time = _frame_to_stamp(step, vstream)

    gop_size = (
        vstream.codec_context.gop_size
    )  # gop size is distance (in frames) between 2 I-frames
    if step > gop_size:
        step_seeking = True
    else:
        step_seeking = False

    seekable = True
    try:
        container.seek(seek_target, stream=vstream)
    except av.error.PermissionError:
        seekable = False
        container, vstream = _open_video(file)  # try again but this time don't seek
        warning(
            "Cannot seek. Possibly a corrupted video file. Retrying with seeking disabled..."
        )

    i = 0
    for packet in container.demux(video=0):
        for frame in packet.decode():
            if packet.pts and packet.pts >= seek_target:
                arr = frame.to_ndarray(format="rgb24")
                video[i] = arr
                i += 1
                seek_target += step_time
                if step_seeking and seekable:
                    container.seek(seek_target, stream=vstream)

        if i == nframes:
            break

    if reverse:
        return video[::-1]
    return video


def _read_timestamps(
    file: Union[str, bytes],
    start: int,
    stop: int,
    step: int,
    reverse: bool,
) -> np.ndarray:
    container, vstream = _open_video(file)

    nframes = math.ceil((stop - start) / step)

    seek_target = _frame_to_stamp(start, vstream)
    step_time = _frame_to_stamp(step, vstream)

    stamps = []
    if vstream.duration is None:
        time_base = 1 / av.time_base
    else:
        time_base = vstream.time_base.numerator / vstream.time_base.denominator

    gop_size = (
        vstream.codec_context.gop_size
    )  # gop size is distance (in frames) between 2 I-frames
    if step > gop_size:
        step_seeking = True
    else:
        step_seeking = False

    seekable = True
    try:
        container.seek(seek_target, stream=vstream)
    except av.error.PermissionError:
        seekable = False
        container, vstream = _open_video(file)  # try again but this time don't seek
        warning(
            "Cannot seek. Possibly a corrupted video file. Retrying with seeking disabled..."
        )

    i = 0
    for packet in container.demux(video=0):
        pts = packet.pts
        if pts and pts >= seek_target:
            stamps.append(pts * time_base)
            i += 1
            seek_target += step_time
            if step_seeking and seekable:
                container.seek(seek_target, stream=vstream)

        if i == nframes:
            break

    # need to sort because when demuxing, frames are in order of dts (decoder timestamp)
    # we need it in order of pts (presentation timestamp)
    stamps.sort()
    stamps_arr = np.zeros((nframes,), dtype=np.float32)
    stamps_arr[: len(stamps)] = stamps

    if reverse:
        return stamps_arr[::-1]
    return stamps_arr


def _open_audio(file: Union[str, bytes, memoryview]):
    if not _PYAV_INSTALLED:
        raise ModuleNotFoundError(
            "PyAV is not installed. Please run `pip install hub[audio]`"
        )
    if isinstance(file, str):
        container = av.open(
            file, options={"protocol_whitelist": "file,http,https,tcp,tls,subfile"}
        )
    else:
        container = av.open(BytesIO(file))

    astreams = container.streams.audio

    if len(astreams) == 0:
        raise IndexError("No audio streams available!")

    astream = astreams[0]

    return container, astream


def _read_shape_from_astream(container, astream):
    nchannels = astream.channels
    duration = astream.duration
    if duration is None:
        duration = container.duration
        if duration is None:
            return (0, nchannels)
        time_base = 1 / av.time_base
    else:
        time_base = astream.time_base.numerator / astream.time_base.denominator
    sample_rate = astream.sample_rate
    nsamples = math.floor(sample_rate * duration * time_base)

    # possible for some files with bad meta
    if nsamples < 0:
        nsamples = 0
    return (nsamples, nchannels)


def _read_audio_shape(
    file: Union[bytes, memoryview, str],
) -> Tuple[int, ...]:
    container, astream = _open_audio(file)
    shape = _read_shape_from_astream(container, astream)
    return shape


def _read_audio_meta(
    file: Union[bytes, memoryview, str],
) -> dict:
    container, astream = _open_audio(file)
    meta = {}
    if astream.duration:
        meta["duration"] = astream.duration
        meta["time_base"] = astream.time_base.numerator / astream.time_base.denominator
    else:
        meta["duration"] = container.duration
        meta["time_base"] = 1 / av.time_base
    meta["sample_rate"] = astream.sample_rate
    meta["duration"] = astream.duration or container.duration
    meta["frame_size"] = astream.frame_size
    meta["nchannels"] = astream.channels
    meta["sample_format"] = astream.format.name
    return meta


def _decompress_audio(
    file: Union[bytes, memoryview, str],
):
    container, astream = _open_audio(file)
    shape = _read_shape_from_astream(container, astream)

    if shape[0] == 0:
        audio = None
        for frame in container.decode(audio=0):
            if not frame.is_corrupt:
                audio = frame.to_ndarray().astype("<f4")
                break

        if audio is not None:
            for frame in container.decode(audio=0):
                if not frame.is_corrupt:
                    audio = np.concatenate(
                        (audio, frame.to_ndarray().astype("<f4")), axis=1
                    )

        return np.transpose(audio)

    audio = np.zeros(shape, dtype="<f4")
    sample_count = 0

    for frame in container.decode(audio=0):
        if not frame.is_corrupt:
            audio[sample_count : sample_count + frame.samples] = (
                frame.to_ndarray().transpose().astype("<f4")
            )
            sample_count += frame.samples
    return audio


def _open_lidar_file(file):
    try:
        import laspy as lp  # type: ignore
    except:
        raise ModuleNotFoundError("laspy not found. Install using `pip install laspy`")
    return lp.read(file)


def _load_lidar_point_cloud_data(file):
    point_cloud = _open_lidar_file(file)
    dimension_names = list(point_cloud.point_format.dimension_names)
    return point_cloud, dimension_names


def _open_point_cloud_data(file: Union[bytes, memoryview, str]):
    if isinstance(file, str):
        point_cloud, dimension_names = _load_lidar_point_cloud_data(file)
        return point_cloud, dimension_names

    point_cloud, dimension_names = _load_lidar_point_cloud_data(BytesIO(file))
    return point_cloud, dimension_names


def _read_point_cloud_meta(file):
    point_cloud, dimension_names = _open_point_cloud_data(file)
    meta_data = {
        "dimension_names": dimension_names,
    }
    if type(point_cloud) != np.ndarray:
        meta_data.update(
            {
                "las_header": {
                    "DEFAULT_VERSION": LAS_HEADER_FILED_NAME_TO_PARSER[
                        "DEFAULT_VERSION"
                    ](point_cloud),
                    "file_source_id": point_cloud.header.file_source_id,
                    "system_identifier": point_cloud.header.system_identifier,
                    "generating_software": point_cloud.header.generating_software,
                    "creation_date": LAS_HEADER_FILED_NAME_TO_PARSER["creation_date"](
                        point_cloud
                    ),
                    "point_count": point_cloud.header.point_count,
                    "scales": point_cloud.header.scales.tolist(),
                    "offsets": point_cloud.header.offsets.tolist(),
                    "number_of_points_by_return": point_cloud.header.number_of_points_by_return.tolist(),
                    "start_of_waveform_data_packet_record": point_cloud.header.start_of_waveform_data_packet_record,
                    "start_of_first_evlr": point_cloud.header.start_of_first_evlr,
                    "number_of_evlrs": point_cloud.header.number_of_evlrs,
                    "version": LAS_HEADER_FILED_NAME_TO_PARSER["version"](point_cloud),
                    "maxs": point_cloud.header.maxs.tolist(),
                    "mins": point_cloud.header.mins.tolist(),
                    "major_version": point_cloud.header.major_version,
                    "minor_version": point_cloud.header.minor_version,
                    "global_encoding": LAS_HEADER_FILED_NAME_TO_PARSER[
                        "global_encoding"
                    ](point_cloud),
                    "uuid": str(point_cloud.header.uuid),
                },
                "vlrs": point_cloud.vlrs,
            }
        )
    return meta_data


def _read_point_cloud_shape_and_dtype(file):
    point_cloud = _decompress_full_point_cloud(file)
    shape = point_cloud.shape
    return shape, point_cloud.dtype


def _decompress_full_point_cloud(file: Union[bytes, memoryview, str]):
    decompressed_point_cloud, _ = _open_point_cloud_data(file)
    meta = _read_point_cloud_meta(file)

    decompressed_point_cloud = np.concatenate(
        [
            np.expand_dims(decompressed_point_cloud[dim_name], -1)
            for dim_name in meta["dimension_names"]
        ],
        axis=1,
    )
    decompressed_point_cloud = decompressed_point_cloud.astype(np.float32)
    return decompressed_point_cloud



import os
import sys
import numpy as np
import pathlib
import pytest
import hub
from hub.core.dataset import Dataset
from hub.core.tensor import Tensor

from hub.tests.common import assert_array_lists_equal, is_opt_true, get_dummy_data_path
from hub.tests.storage_fixtures import enabled_remote_storages
from hub.core.storage import GCSProvider
from hub.util.exceptions import (
    InvalidOperationError,
    TensorDtypeMismatchError,
    TensorDoesNotExistError,
    TensorAlreadyExistsError,
    TensorGroupDoesNotExistError,
    TensorGroupAlreadyExistsError,
    TensorInvalidSampleShapeError,
    DatasetHandlerError,
    UnsupportedCompressionError,
    InvalidTensorNameError,
    InvalidTensorGroupNameError,
    RenameError,
    PathNotEmptyException,
    BadRequestException,
    ReadOnlyModeError,
    EmptyTensorError,
    InvalidTokenException,
    TokenPermissionError,
    UserNotLoggedInException,
    SampleAppendingError,
)
from hub.util.path import convert_string_to_pathlib_if_needed
from hub.util.pretty_print import summary_tensor, summary_dataset
from hub.constants import GDRIVE_OPT, MB
from hub.client.config import REPORTING_CONFIG_FILE_PATH

from click.testing import CliRunner
from hub.cli.auth import login, logout
from hub.util.bugout_reporter import feature_report_path


# need this for 32-bit and 64-bit systems to have correct tests
MAX_INT_DTYPE = np.int_.__name__
MAX_FLOAT_DTYPE = np.float_.__name__


# not using the predefined parametrizes because `hub_cloud_ds_generator` is not enabled by default
@pytest.mark.parametrize(
    "ds_generator",
    [
        "local_ds_generator",
        "s3_ds_generator",
        "hub_cloud_ds_generator",
    ],
    indirect=True,
)
def test_persist(ds_generator):
    ds = ds_generator()

    ds.create_tensor("image")
    ds.image.extend(np.ones((4, 224, 224, 3)))

    ds_new = ds_generator()
    assert len(ds_new) == 4

    assert ds_new.image.shape == (4, 224, 224, 3)
    np.testing.assert_array_equal(ds_new.image.numpy(), np.ones((4, 224, 224, 3)))

    assert ds_new.meta.version == hub.__version__

    ds_new.create_tensor("label")
    ds_new.label.extend([1, 2, 3, 4])

    ds2 = ds_generator()

    ds2.storage["dataset_meta.json"] == ds_new.storage["dataset_meta.json"]
    assert len(ds2) == 4
    np.testing.assert_array_equal(ds2.label.numpy(), np.array([[1], [2], [3], [4]]))


def test_persist_keys(local_ds_generator):
    ds = local_ds_generator()

    ds.create_tensor("image")

    ds_new = local_ds_generator()
    assert set(ds_new.storage.keys()) == {
        "dataset_meta.json",
        "image/commit_diff",
        "image/tensor_meta.json",
        "_image_id/tensor_meta.json",
        "_image_id/commit_diff",
        "_image_shape/tensor_meta.json",
        "_image_shape/commit_diff",
    }


def test_persist_with(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("image")
        ds.image.extend(np.ones((4, 224, 224, 3)))

        ds_new = local_ds_generator()
        assert len(ds_new) == 0  # shouldn't be flushed yet

    ds_new = local_ds_generator()
    assert len(ds_new) == 4

    engine = ds_new.image.chunk_engine
    assert engine.chunk_id_encoder.num_samples == ds_new.image.meta.length
    assert engine.chunk_id_encoder.num_chunks == 1

    assert ds_new.image.shape == (4, 224, 224, 3)

    np.testing.assert_array_equal(ds_new.image.numpy(), np.ones((4, 224, 224, 3)))

    assert ds_new.meta.version == hub.__version__


def test_persist_clear_cache(local_ds_generator):
    ds = local_ds_generator()
    ds.create_tensor("image")
    ds.image.extend(np.ones((4, 224, 224, 3)))
    ds.clear_cache()
    ds_new = local_ds_generator()
    assert len(ds_new) == 4

    assert ds_new.image.shape == (4, 224, 224, 3)

    np.testing.assert_array_equal(ds_new.image.numpy(), np.ones((4, 224, 224, 3)))


def test_populate_dataset(local_ds):
    assert local_ds.meta.tensors == []
    local_ds.create_tensor("image")
    assert len(local_ds) == 0
    assert len(local_ds.image) == 0

    local_ds.image.extend(np.ones((4, 28, 28)))
    assert len(local_ds) == 4
    assert len(local_ds.image) == 4

    for _ in range(10):
        local_ds.image.append(np.ones((28, 28)))
    assert len(local_ds.image) == 14

    local_ds.image.extend([np.ones((28, 28)), np.ones((28, 28))])
    assert len(local_ds.image) == 16
    assert len(local_ds.image.numpy()) == 16
    assert len(local_ds.image[0:5].numpy()) == 5
    assert len(local_ds.image[:-1].numpy()) == 15
    assert len(local_ds.image[-5:].numpy()) == 5

    assert local_ds.meta.tensors == ["image", "_image_shape", "_image_id"]
    assert local_ds.meta.version == hub.__version__


def test_larger_data_memory(memory_ds):
    memory_ds.create_tensor("image", max_chunk_size=2 * MB)
    x = np.ones((4, 1024, 1024))
    memory_ds.image.extend(x)
    assert len(memory_ds) == 4
    assert memory_ds.image.shape == x.shape
    np.testing.assert_array_equal(memory_ds.image.numpy(), x)
    idxs = [
        0,
        1,
        3,
        -1,
        slice(0, 3),
        slice(2, 4),
        slice(2, None),
        (0, slice(5, None), slice(None, 714)),
        (2, 100, 1007),
        (slice(1, 3), [20, 1000, 2, 400], [-2, 3, 577, 1023]),
    ]
    for idx in idxs:
        np.testing.assert_array_equal(memory_ds.image[idx].numpy(), x[idx])


def test_stringify(memory_ds, capsys):
    ds = memory_ds
    ds.create_tensor("image")
    ds.image.extend(np.ones((4, 4)))

    ds.summary()
    assert (
        capsys.readouterr().out
        == "Dataset(path='mem://hub_pytest/test_api/test_stringify', tensors=['image'])\n\n tensor    htype    shape    dtype  compression\n -------  -------  -------  -------  ------- \n  image   generic  (4, 4)    None     None   \n"
    )
    ds[1:2].summary()
    assert (
        capsys.readouterr().out
        == "Dataset(path='mem://hub_pytest/test_api/test_stringify', index=Index([slice(1, 2, None)]), tensors=['image'])\n\n tensor    htype    shape    dtype  compression\n -------  -------  -------  -------  ------- \n  image   generic  (1, 4)    None     None   \n"
    )
    ds.image.summary()
    assert (
        capsys.readouterr().out
        == "Tensor(key='image')\n\n  htype    shape    dtype  compression\n -------  -------  -------  ------- \n generic  (4, 4)    None     None   \n"
    )
    ds[1:2].image.summary()
    assert (
        capsys.readouterr().out
        == "Tensor(key='image', index=Index([slice(1, 2, None)]))\n\n  htype    shape    dtype  compression\n -------  -------  -------  ------- \n generic  (1, 4)    None     None   \n"
    )


def test_summary(memory_ds):
    ds = memory_ds
    ds.create_tensor("abc")
    ds.abc.extend(np.ones((4, 4)))
    ds.create_tensor("images", htype="image", dtype="int32", sample_compression="jpeg")

    assert (
        summary_dataset(ds)
        == "\n tensor    htype    shape    dtype  compression\n -------  -------  -------  -------  ------- \n   abc    generic  (4, 4)    None     None   \n images    image    (0,)     int32    jpeg   "
    )
    assert (
        summary_tensor(ds.abc)
        == "\n  htype    shape    dtype  compression\n -------  -------  -------  ------- \n generic  (4, 4)    None     None   "
    )
    assert (
        summary_tensor(ds.images)
        == "\n  htype    shape    dtype  compression\n -------  -------  -------  ------- \n  image    (0,)     int32    jpeg   "
    )


def test_stringify_with_path(local_ds, capsys):
    ds = local_ds
    assert local_ds.path
    ds.summary()
    assert (
        capsys.readouterr().out
        == f"Dataset(path='{local_ds.path}', tensors=[])\n\n tensor    htype    shape    dtype  compression\n -------  -------  -------  -------  ------- \n"
    )


def test_fixed_tensor(local_ds):
    local_ds.create_tensor("image")
    local_ds.image.extend(np.ones((32, 28, 28)))
    assert len(local_ds) == 32
    np.testing.assert_array_equal(local_ds.image.numpy(), np.ones((32, 28, 28)))


def test_dynamic_tensor(local_ds):
    local_ds.create_tensor("image")

    a1 = np.ones((32, 28, 28))
    a2 = np.ones((10, 36, 11))
    a3 = np.ones((29, 10))

    image = local_ds.image

    image.extend(a1)
    image.extend(a2)
    image.append(a3)

    expected_list = [*a1, *a2, a3]
    actual_list = image.numpy(aslist=True)

    assert type(actual_list) == list
    assert_array_lists_equal(expected_list, actual_list)

    # test negative indexing
    np.testing.assert_array_equal(expected_list[-1], image[-1].numpy())
    np.testing.assert_array_equal(expected_list[-2], image[-2].numpy())
    assert_array_lists_equal(expected_list[-2:], image[-2:].numpy(aslist=True))
    assert_array_lists_equal(expected_list[::-3], image[::-3].numpy(aslist=True))

    assert image.shape == (43, None, None)
    assert image.shape_interval.lower == (43, 28, 10)
    assert image.shape_interval.upper == (43, 36, 28)
    assert image.is_dynamic


def test_empty_samples(local_ds: Dataset):
    tensor = local_ds.create_tensor("with_empty")
    with pytest.raises(EmptyTensorError):
        ds_pytorch = local_ds.pytorch()

    a1 = np.arange(25 * 4 * 2).reshape(25, 4, 2)
    a2 = np.arange(5 * 10 * 50 * 2).reshape(5, 10, 50, 2)
    a3 = np.arange(0).reshape(0, 0, 2)
    a4 = np.arange(0).reshape(9, 0, 10, 2)

    tensor.append(a1)
    tensor.extend(a2)
    tensor.append(a3)
    tensor.extend(a4)

    actual_list = tensor.numpy(aslist=True)
    expected_list = [a1, *a2, a3, *a4]

    assert tensor.meta.sample_compression is None

    assert len(tensor) == 16
    assert tensor.shape_interval.lower == (16, 0, 0, 2)
    assert tensor.shape_interval.upper == (16, 25, 50, 2)

    assert_array_lists_equal(actual_list, expected_list)

    # test indexing individual empty samples with numpy while looping, this may seem redundant but this was failing before
    for actual_sample, expected in zip(local_ds, expected_list):
        actual = actual_sample.with_empty.numpy()
        np.testing.assert_array_equal(actual, expected)


def test_indexed_tensor(local_ds: Dataset):
    tensor = local_ds.create_tensor("abc")
    tensor.append([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
    np.testing.assert_array_equal(tensor[0, 1].numpy(), np.array([4, 5, 6]))
    np.testing.assert_array_equal(
        tensor[0, 0:2].numpy(), np.array([[1, 2, 3], [4, 5, 6]])
    )
    np.testing.assert_array_equal(
        tensor[0, 0::2].numpy(), np.array([[1, 2, 3], [7, 8, 9]])
    )


def test_safe_downcasting(local_ds):
    int_tensor = local_ds.create_tensor("int", dtype="uint8")
    int_tensor.append(0)
    int_tensor.append(1)
    int_tensor.extend([2, 3, 4])
    int_tensor.extend([5, 6, np.uint8(7)])
    int_tensor.append(np.zeros((0,), dtype="uint64"))
    with pytest.raises(TensorDtypeMismatchError):
        int_tensor.append(-8)
    int_tensor.append(np.array([1]))
    assert len(int_tensor) == 10
    with pytest.raises(TensorDtypeMismatchError):
        int_tensor.append(np.array([1.0]))

    float_tensor = local_ds.create_tensor("float", dtype="float32")
    float_tensor.append(0)
    float_tensor.append(1)
    float_tensor.extend([2, 3.0, 4.0])
    float_tensor.extend([5.0, 6.0, np.float32(7.0)])
    with pytest.raises(TensorDtypeMismatchError):
        float_tensor.append(float(np.finfo(np.float32).max + 1))
    float_tensor.append(np.array([1]))
    float_tensor.append(np.array([1.0]))
    assert len(float_tensor) == 10


def test_scalar_samples(local_ds):
    tensor = local_ds.create_tensor("scalars")

    assert tensor.meta.dtype is None

    # first sample sets dtype
    tensor.append(5)
    assert tensor.meta.dtype == MAX_INT_DTYPE

    with pytest.raises(TensorDtypeMismatchError):
        tensor.append(5.1)

    tensor.append(10)
    tensor.append(-99)
    tensor.append(np.array(4))

    tensor.append(np.int16(4))

    with pytest.raises(TensorDtypeMismatchError):
        tensor.append(np.float32(4))

    tensor.append(np.uint8(3))

    tensor.extend([10, 1, 4])
    tensor.extend([1])
    tensor.extend(np.array([1, 2, 3], dtype=MAX_INT_DTYPE))

    tensor.extend(np.array([4, 5, 33], dtype="int16"))

    assert len(tensor) == 16

    assert tensor.shape == (16, 1)

    tensor.append([1])
    tensor.append([1, 2, 3])
    tensor.extend([[1], [2], [3, 4]])
    tensor.append(np.empty(0, dtype=int))

    with pytest.raises(TensorInvalidSampleShapeError):
        tensor.append([[[1]]])

    expected = [
        [5],
        [10],
        [-99],
        [4],
        [4],
        [3],
        [10],
        [1],
        [4],
        [1],
        [1],
        [2],
        [3],
        [4],
        [5],
        [33],
        [1],
        [1, 2, 3],
        [1],
        [2],
        [3, 4],
        [],
    ]

    assert_array_lists_equal(expected, tensor.numpy(aslist=True))

    assert tensor.shape == (22, None)
    assert tensor.shape_interval.lower == (22, 0)
    assert tensor.shape_interval.upper == (22, 3)

    assert len(tensor) == 22


def test_sequence_samples(local_ds):
    tensor = local_ds.create_tensor("arrays")

    tensor.append([1, 2, 3])
    tensor.extend([[4, 5, 6]])
    local_ds.clear_cache()

    assert len(tensor) == 2
    expected_list = [[1, 2, 3], [4, 5, 6]]
    expected = np.array(expected_list)
    np.testing.assert_array_equal(tensor.numpy(), expected)

    assert type(tensor.numpy(aslist=True)) == list
    assert_array_lists_equal(tensor.numpy(aslist=True), expected_list)


def test_iterate_dataset(local_ds):
    labels = [1, 9, 7, 4]
    local_ds.create_tensor("image")
    local_ds.create_tensor("label")

    local_ds.image.extend(np.ones((4, 28, 28)))
    local_ds.label.extend(np.asarray(labels).reshape((4, 1)))

    for idx, sub_ds in enumerate(local_ds):
        img = sub_ds.image.numpy()
        label = sub_ds.label.numpy()
        np.testing.assert_array_equal(img, np.ones((28, 28)))
        assert label.shape == (1,)
        assert label == labels[idx]


def _check_tensor(tensor, data):
    np.testing.assert_array_equal(tensor.numpy(), data)


def test_compute_slices(memory_ds):
    ds = memory_ds
    shape = (64, 16, 16, 16)
    data = np.arange(np.prod(shape)).reshape(shape)
    ds.create_tensor("data")
    ds.data.extend(data)

    _check_tensor(ds.data[:], data[:])
    _check_tensor(ds.data[10:20], data[10:20])
    _check_tensor(ds.data[5], data[5])
    _check_tensor(ds.data[0][:], data[0][:])
    _check_tensor(ds.data[-1][:], data[-1][:])
    _check_tensor(ds.data[3, 3], data[3, 3])
    _check_tensor(ds.data[30:40, :, 8:11, 4], data[30:40, :, 8:11, 4])
    _check_tensor(ds.data[16, 4, 5, 1:3], data[16, 4, 5, 1:3])
    _check_tensor(ds[[0, 1, 2, 5, 6, 10, 60]].data, data[[0, 1, 2, 5, 6, 10, 60]])
    _check_tensor(ds.data[[0, 1, -2, 5, -6, 10, 60]], data[[0, 1, -2, 5, -6, 10, 60]])
    _check_tensor(ds.data[0][[0, 1, 2, 5, 6, 10, 15]], data[0][[0, 1, 2, 5, 6, 10, 15]])
    _check_tensor(ds.data[[3, 2, 1, 0]][0], data[[3, 2, 1, 0]][0])
    _check_tensor(ds[[3, 2, 1, 0]][0].data, data[[3, 2, 1, 0]][0])
    _check_tensor(ds[[3, 2, 1, 0]].data[0], data[3])
    _check_tensor(ds[(0, 1, 6, 10, 15), :].data, data[(0, 1, 6, 10, 15), :])
    _check_tensor(ds.data[(0, 1, 6, 10, 15), :], data[(0, 1, 6, 10, 15), :])
    _check_tensor(ds.data[0][(0, 1, 6, 10, 15), :], data[0][(0, 1, 6, 10, 15), :])
    _check_tensor(ds.data[0, (0, 1, 5)], data[0, (0, 1, 5)])
    _check_tensor(ds.data[:, :][0], data[:, :][0])
    _check_tensor(ds.data[:, :][0:2], data[:, :][0:2])
    _check_tensor(ds.data[0, :][0:2], data[0, :][0:2])
    _check_tensor(ds.data[:, 0][0:2], data[:, 0][0:2])
    _check_tensor(ds.data[:, 0][0:2], data[:, 0][0:2])
    _check_tensor(ds.data[:, :][0][(0, 1, 2), 0], data[:, :][0][(0, 1, 2), 0])
    _check_tensor(ds.data[0][(0, 1, 2), 0][1], data[0][(0, 1, 2), 0][1])
    _check_tensor(ds.data[:, :][0][(0, 1, 2), 0][1], data[:, :][0][(0, 1, 2), 0][1])
    _check_tensor(ds.data[::-1], data[::-1])
    _check_tensor(ds.data[::-3], data[::-3])
    _check_tensor(ds.data[::-3][4], data[::-3][4])
    _check_tensor(ds.data[-2:], data[-2:])
    _check_tensor(ds.data[-6:][3], data[-6:][3])
    _check_tensor(ds.data[:-6:][3], data[:-6:][3])


def test_length_slices(memory_ds):
    ds = memory_ds
    data = np.array([1, 2, 3, 9, 8, 7, 100, 99, 98, 99, 101])
    data_2 = np.array([1, 2, 3, 9, 8, 7, 100, 99, 98, 99, 101, 12, 15, 18])
    ds.create_tensor("data")
    ds.create_tensor("data_2")

    ds.data.extend(data)
    ds.data_2.extend(data_2)

    assert len(ds) == 11
    assert ds.min_len == len(ds)
    assert ds.max_len == 14
    assert len(ds[0]) == 1
    assert len(ds[0:1]) == 1
    assert len(ds[0:0]) == 0
    assert len(ds[1:10]) == 9
    assert len(ds[1:7:2]) == 3
    assert len(ds[1:8:2]) == 4
    assert len(ds[1:9:2]) == 4
    assert len(ds[1:10:2]) == 5
    assert len(ds[[0, 1, 5, 9]]) == 4

    assert len(ds.data) == 11
    assert len(ds.data[0]) == 1
    assert len(ds.data[0:1]) == 1
    assert len(ds.data[0:0]) == 0
    assert len(ds.data[1:10]) == 9
    assert len(ds.data[1:7:2]) == 3
    assert len(ds.data[1:8:2]) == 4
    assert len(ds.data[1:9:2]) == 4
    assert len(ds.data[1:10:2]) == 5
    assert len(ds.data[[0, 1, 5, 9]]) == 4

    assert ds.data.shape == (11, 1)
    assert ds[0:5].data.shape == (5, 1)
    assert ds.data[1:6].shape == (5, 1)


def test_shape_property(memory_ds):
    fixed = memory_ds.create_tensor("fixed_tensor")
    dynamic = memory_ds.create_tensor("dynamic_tensor")

    # dynamic shape property
    dynamic.extend(np.ones((32, 28, 20, 2)))
    dynamic.extend(np.ones((16, 33, 20, 5)))
    assert dynamic.shape == (48, None, 20, None)
    assert dynamic.shape_interval.lower == (48, 28, 20, 2)
    assert dynamic.shape_interval.upper == (48, 33, 20, 5)
    assert dynamic.is_dynamic

    # fixed shape property
    fixed.extend(np.ones((9, 28, 28)))
    fixed.extend(np.ones((13, 28, 28)))
    assert fixed.shape == (22, 28, 28)
    assert fixed.shape_interval.lower == (22, 28, 28)
    assert fixed.shape_interval.upper == (22, 28, 28)
    assert not fixed.is_dynamic


def test_htype(memory_ds: Dataset):
    image = memory_ds.create_tensor("image", htype="image", sample_compression="png")
    bbox = memory_ds.create_tensor("bbox", htype="bbox")
    label = memory_ds.create_tensor("label", htype="class_label")
    video = memory_ds.create_tensor("video", htype="video", sample_compression="mkv")
    bin_mask = memory_ds.create_tensor("bin_mask", htype="binary_mask")
    segment_mask = memory_ds.create_tensor("segment_mask", htype="segment_mask")
    keypoints_coco = memory_ds.create_tensor("keypoints_coco", htype="keypoints_coco")
    point = memory_ds.create_tensor("point", htype="point")
    point_cloud = memory_ds.create_tensor(
        "point_cloud", htype="point_cloud", sample_compression="las"
    )
    memory_ds.create_tensor(
        "point_cloud_calibration_matrix", htype="point_cloud.calibration_matrix"
    )

    image.append(np.ones((28, 28, 3), dtype=np.uint8))
    bbox.append(np.array([1.0, 1.0, 0.0, 0.5], dtype=np.float32))
    # label.append(5)
    label.append(np.array(5, dtype=np.uint32))
    with pytest.raises(NotImplementedError):
        video.append(np.ones((10, 28, 28, 3), dtype=np.uint8))
    bin_mask.append(np.zeros((28, 28), dtype=np.bool8))
    segment_mask.append(np.ones((28, 28), dtype=np.uint32))
    keypoints_coco.append(np.ones((51, 2), dtype=np.int32))
    point.append(np.ones((11, 2), dtype=np.int32))

    point_cloud.append(
        hub.read(os.path.join(get_dummy_data_path("point_cloud"), "point_cloud.las"))
    )
    point_cloud_dummy_data_path = pathlib.Path(get_dummy_data_path("point_cloud"))
    point_cloud.append(hub.read(point_cloud_dummy_data_path / "point_cloud.las"))
    # Along the forst direcection three matrices are concatenated, the first matrix is P,
    # the second one is Tr and the third one is R
    memory_ds.point_cloud_calibration_matrix.append(
        np.zeros((3, 4, 4), dtype=np.float32)
    )


def test_dtype(memory_ds: Dataset):
    tensor = memory_ds.create_tensor("tensor")
    dtyped_tensor = memory_ds.create_tensor("dtyped_tensor", dtype="uint8")
    np_dtyped_tensor = memory_ds.create_tensor(
        "np_dtyped_tensor", dtype=MAX_FLOAT_DTYPE
    )
    py_dtyped_tensor = memory_ds.create_tensor("py_dtyped_tensor", dtype=float)

    # .meta.dtype should always be str or None
    assert type(tensor.meta.dtype) == type(None)
    assert type(dtyped_tensor.meta.dtype) == str
    assert type(np_dtyped_tensor.meta.dtype) == str
    assert type(py_dtyped_tensor.meta.dtype) == str

    # .dtype should always be np.dtype or None
    assert type(tensor.dtype) == type(
        None
    ), "An htype with a generic `dtype` should start as None... If this check doesn't exist, float64/float32 may be it's initial type."
    assert dtyped_tensor.dtype == np.uint8
    assert np_dtyped_tensor.dtype == MAX_FLOAT_DTYPE
    assert py_dtyped_tensor.dtype == MAX_FLOAT_DTYPE

    tensor.append(np.ones((10, 10), dtype="float32"))
    dtyped_tensor.append(np.ones((10, 10), dtype="uint8"))
    np_dtyped_tensor.append(np.ones((10, 10), dtype=MAX_FLOAT_DTYPE))
    py_dtyped_tensor.append(np.ones((10, 10), dtype=MAX_FLOAT_DTYPE))

    # test auto upcasting
    np_dtyped_tensor.append(np.ones((10, 10), dtype="float32"))
    py_dtyped_tensor.append(np.ones((10, 10), dtype="float32"))

    with pytest.raises(TensorDtypeMismatchError):
        tensor.append(np.ones((10, 10), dtype="float64"))

    with pytest.raises(TensorDtypeMismatchError):
        dtyped_tensor.append(np.ones((10, 10), dtype="uint64") * 256)

    assert tensor.dtype == np.float32
    assert dtyped_tensor.dtype == np.uint8
    assert np_dtyped_tensor.dtype == MAX_FLOAT_DTYPE
    assert py_dtyped_tensor.dtype == MAX_FLOAT_DTYPE

    assert len(tensor) == 1
    assert len(dtyped_tensor) == 1


@pytest.mark.xfail(raises=TypeError, strict=True)
def test_fails_on_wrong_tensor_syntax(memory_ds):
    memory_ds.some_tensor = np.ones((28, 28))


def test_array_interface(memory_ds: Dataset):
    tensor = memory_ds.create_tensor("tensor")
    x = np.arange(10).reshape(5, 2)
    tensor.append(x)
    arr1 = np.array(tensor)
    arr2 = np.array(tensor)
    np.testing.assert_array_equal(x, arr1[0])
    np.testing.assert_array_equal(x, arr2[0])
    tensor.append(x)
    np.testing.assert_array_equal(tensor.numpy(), np.concatenate([arr1, arr2]))


def test_hub_dataset_suffix_bug(hub_cloud_ds, hub_cloud_dev_token):
    # creating dataset with similar name but some suffix removed from end
    ds = hub.dataset(hub_cloud_ds.path[:-1], token=hub_cloud_dev_token)

    # need to delete because it's a different path (won't be auto cleaned up)
    ds.delete()


def test_index_range(memory_ds):
    with pytest.raises(IndexError):
        memory_ds[0]

    memory_ds.create_tensor("label")

    with pytest.raises(IndexError):
        memory_ds.label[0]

    memory_ds.label.extend([5, 6, 7])
    assert len(memory_ds) == 3
    assert len(memory_ds.label) == 3

    for valid_idx in [0, 1, 2, -0, -1, -2, -3]:
        memory_ds[valid_idx]
        memory_ds.label[valid_idx]

    for invalid_idx in [3, 4, -4, -5]:
        with pytest.raises(IndexError):
            memory_ds[invalid_idx]
        with pytest.raises(IndexError):
            memory_ds.label[invalid_idx]

    memory_ds[[0, 1, 2]]
    with pytest.raises(IndexError):
        memory_ds[[0, 1, 2, 3, 4, 5]]


@pytest.mark.parametrize("convert_to_pathlib", [True, False])
def test_empty_dataset(convert_to_pathlib):
    with CliRunner().isolated_filesystem():
        test_path = "test"
        test_path = convert_string_to_pathlib_if_needed(test_path, convert_to_pathlib)
        ds = hub.dataset(test_path)
        ds.create_tensor("x")
        ds.create_tensor("y")
        ds.create_tensor("z")
        ds = hub.dataset(test_path)
        assert list(ds.tensors) == ["x", "y", "z"]


@pytest.mark.parametrize("convert_to_pathlib", [True, False])
def test_like(local_path, convert_to_pathlib):
    src_path = os.path.join(local_path, "src")
    src_path = convert_string_to_pathlib_if_needed(src_path, convert_to_pathlib)
    dest_path = os.path.join(local_path, "dest")
    dest_path = convert_string_to_pathlib_if_needed(dest_path, convert_to_pathlib)

    src_ds = hub.dataset(src_path)
    src_ds.info.update(key=0)

    src_ds.create_tensor("a", htype="image", sample_compression="png")
    src_ds.create_tensor("b", htype="class_label")
    src_ds.create_tensor("c")
    src_ds.create_tensor("d", dtype=bool)

    src_ds.d.info.update(key=1)

    assert src_ds.info.key == 0
    assert src_ds.d.info.key == 1

    dest_ds = hub.like(dest_path, src_ds)

    assert tuple(dest_ds.tensors.keys()) == ("a", "b", "c", "d")

    assert dest_ds.a.meta.htype == "image"
    assert dest_ds.a.meta.sample_compression == "png"
    assert dest_ds.b.meta.htype == "class_label"
    assert dest_ds.c.meta.htype is None
    assert dest_ds.d.dtype == bool

    assert dest_ds.info.key == 0
    assert dest_ds.d.info.key == 1

    assert len(dest_ds) == 0


def test_tensor_creation_fail_recovery():
    with CliRunner().isolated_filesystem():
        ds = hub.dataset("test")
        with ds:
            ds.create_tensor("x")
            ds.create_tensor("y")
            with pytest.raises(UnsupportedCompressionError):
                ds.create_tensor("z", sample_compression="something_random")
        ds = hub.dataset("test")
        assert list(ds.tensors) == ["x", "y"]
        ds.create_tensor("z")
        assert list(ds.tensors) == ["x", "y", "z"]


def test_dataset_delete():
    with CliRunner().isolated_filesystem():
        os.mkdir("test")
        with open("test/test.txt", "w") as f:
            f.write("some data")

        with pytest.raises(DatasetHandlerError):
            # Can't delete raw data without force
            hub.delete("test/")

        hub.delete("test/", force=True)
        assert not os.path.isfile("test/test.txt")

        hub.empty("test/").create_tensor("tmp")
        assert os.path.isfile("test/dataset_meta.json")

        hub.delete("test/")
        assert not os.path.isfile("test/dataset_meta.json")

        pathlib_path = pathlib.Path("test/")
        hub.empty(pathlib_path).create_tensor("tmp")
        assert os.path.isfile("test/dataset_meta.json")

        hub.delete(pathlib_path)
        assert not os.path.isfile("test/dataset_meta.json")

        old_size = hub.constants.DELETE_SAFETY_SIZE
        hub.constants.DELETE_SAFETY_SIZE = 1 * MB

        ds = hub.empty("test/")
        ds.create_tensor("data")
        ds.data.extend(np.zeros((100, 2000)))

        try:
            hub.delete("test/")
        finally:
            assert os.path.isfile("test/dataset_meta.json")

        hub.delete("test/", large_ok=True)
        assert not os.path.isfile("test/dataset_meta.json")

        hub.constants.DELETE_SAFETY_SIZE = old_size


@pytest.mark.parametrize(
    ("ds_generator", "path", "hub_token"),
    [
        ("local_ds_generator", "local_path", "hub_cloud_dev_token"),
        ("s3_ds_generator", "s3_path", "hub_cloud_dev_token"),
        ("gcs_ds_generator", "gcs_path", "hub_cloud_dev_token"),
        ("hub_cloud_ds_generator", "hub_cloud_path", "hub_cloud_dev_token"),
    ],
    indirect=True,
)
@pytest.mark.parametrize("convert_to_pathlib", [True, False])
def test_dataset_rename(ds_generator, path, hub_token, convert_to_pathlib):
    ds = ds_generator()
    ds.create_tensor("abc")
    ds.abc.append([1, 2, 3, 4])

    new_path = "_".join([path, "renamed"])

    ds.path = convert_string_to_pathlib_if_needed(ds.path, convert_to_pathlib)
    new_path = convert_string_to_pathlib_if_needed(new_path, convert_to_pathlib)

    with pytest.raises(RenameError):
        ds.rename("wrongfolder/new_ds")

    if str(ds.path).startswith("hub://"):
        with pytest.raises(BadRequestException):
            ds.rename(ds.path)
    else:
        with pytest.raises(PathNotEmptyException):
            ds.rename(ds.path)

    ds = hub.rename(ds.path, new_path, token=hub_token)
    assert ds.path == str(new_path)
    np.testing.assert_array_equal(ds.abc.numpy(), np.array([[1, 2, 3, 4]]))

    ds = hub.load(new_path, token=hub_token)
    np.testing.assert_array_equal(ds.abc.numpy(), np.array([[1, 2, 3, 4]]))

    with pytest.raises(InvalidTokenException):
        ds = hub.load(
            "hub://activeloop-test/sohas-weapons-train", token="invalid token"
        )

    with pytest.raises(InvalidTokenException):
        ds = hub.empty(
            "hub://activeloop-test/sohas-weapons-train", token="invalid token"
        )

    with pytest.raises(InvalidTokenException):
        ds = hub.dataset(
            "hub://activeloop-test/sohas-weapons-train", token="invalid token"
        )

    hub.delete(new_path, token=hub_token)


@pytest.mark.parametrize(
    "path,hub_token",
    [
        ["local_path", "hub_cloud_dev_token"],
        ["hub_cloud_path", "hub_cloud_dev_token"],
    ],
    indirect=True,
)
@pytest.mark.parametrize("num_workers", [0, 2])
@pytest.mark.parametrize("progressbar", [True, False])
def test_dataset_deepcopy(path, hub_token, num_workers, progressbar):
    src_path = "_".join((path, "src"))
    dest_path = "_".join((path, "dest"))

    src_ds = hub.empty(src_path, overwrite=True, token=hub_token)

    with src_ds:
        src_ds.info.update(key=0)

        src_ds.create_tensor("a", htype="image", sample_compression="png")
        src_ds.create_tensor("b", htype="class_label")
        src_ds.create_tensor("c")
        src_ds.create_tensor("d", dtype=bool)

        src_ds.d.info.update(key=1)

        src_ds["a"].append(np.ones((28, 28), dtype="uint8"))
        src_ds["b"].append(0)

    dest_ds = hub.deepcopy(
        src_path,
        dest_path,
        overwrite=True,
        src_token=hub_token,
        dest_token=hub_token,
        num_workers=num_workers,
        progressbar=progressbar,
    )

    assert list(dest_ds.tensors) == ["a", "b", "c", "d"]
    assert dest_ds.a.meta.htype == "image"
    assert dest_ds.a.meta.sample_compression == "png"
    assert dest_ds.b.meta.htype == "class_label"
    assert dest_ds.c.meta.htype == None
    assert dest_ds.d.dtype == bool

    assert dest_ds.info.key == 0
    assert dest_ds.d.info.key == 1

    for tensor in dest_ds.meta.tensors:
        np.testing.assert_array_equal(src_ds[tensor].numpy(), dest_ds[tensor].numpy())

    with pytest.raises(DatasetHandlerError):
        hub.deepcopy(src_path, dest_path, src_token=hub_token, dest_token=hub_token)

    hub.deepcopy(
        src_path,
        dest_path,
        overwrite=True,
        src_token=hub_token,
        dest_token=hub_token,
        num_workers=num_workers,
        progressbar=progressbar,
    )

    assert list(dest_ds.tensors) == ["a", "b", "c", "d"]
    for tensor in dest_ds.tensors:
        np.testing.assert_array_equal(src_ds[tensor].numpy(), dest_ds[tensor].numpy())

    # test fot dataset.load:
    dest_ds = hub.load(dest_path, token=hub_token)
    assert list(dest_ds.tensors) == ["a", "b", "c", "d"]
    for tensor in dest_ds.tensors.keys():
        np.testing.assert_array_equal(src_ds[tensor].numpy(), dest_ds[tensor].numpy())

    hub.deepcopy(
        src_path,
        dest_path,
        overwrite=True,
        src_token=hub_token,
        dest_token=hub_token,
        num_workers=num_workers,
        progressbar=progressbar,
    )
    dest_ds = hub.load(dest_path, token=hub_token)

    assert list(dest_ds.tensors) == ["a", "b", "c", "d"]
    for tensor in dest_ds.tensors:
        np.testing.assert_array_equal(src_ds[tensor].numpy(), dest_ds[tensor].numpy())

    hub.deepcopy(
        src_path,
        dest_path,
        tensors=["a", "d"],
        overwrite=True,
        src_token=hub_token,
        dest_token=hub_token,
        num_workers=num_workers,
        progressbar=progressbar,
    )
    dest_ds = hub.load(dest_path, token=hub_token)
    assert list(dest_ds.tensors) == ["a", "d"]
    for tensor in dest_ds.tensors:
        np.testing.assert_array_equal(src_ds[tensor].numpy(), dest_ds[tensor].numpy())
    hub.delete(src_path, token=hub_token)
    hub.delete(dest_path, token=hub_token)


def test_cloud_delete_doesnt_exist(hub_cloud_path, hub_cloud_dev_token):
    username = hub_cloud_path.split("/")[2]
    # this dataset doesn't exist
    new_path = f"hub://{username}/doesntexist123"
    hub.delete(new_path, token=hub_cloud_dev_token, force=True)


def test_invalid_tensor_name(memory_ds):
    with pytest.raises(InvalidTensorNameError):
        memory_ds.create_tensor("group/version_state")
    with pytest.raises(InvalidTensorNameError):
        memory_ds.create_tensor("info")


def test_compressions_list():
    assert hub.compressions == [
        "apng",
        "avi",
        "bmp",
        "dcm",
        "dib",
        "flac",
        "gif",
        "ico",
        "jpeg",
        "jpeg2000",
        "las",
        "lz4",
        "mkv",
        "mp3",
        "mp4",
        "pcx",
        "png",
        "ppm",
        "sgi",
        "tga",
        "tiff",
        "wav",
        "webp",
        "wmf",
        "xbm",
        None,
    ]


def test_htypes_list():
    assert hub.htypes == [
        "audio",
        "bbox",
        "bbox.3d",
        "binary_mask",
        "class_label",
        "dicom",
        "generic",
        "image",
        "image.gray",
        "image.rgb",
        "instance_label",
        "json",
        "keypoints_coco",
        "list",
        "point",
        "point_cloud",
        "point_cloud.calibration_matrix",
        "segment_mask",
        "text",
        "video",
    ]


def test_groups(local_ds_generator):
    ds = local_ds_generator()
    ds.create_tensor("x")
    with pytest.raises(TensorAlreadyExistsError):
        ds.create_tensor("x/y")
    ds.create_tensor("y/x")
    with pytest.raises(TensorGroupAlreadyExistsError):
        ds.create_tensor("y")
    assert isinstance(ds.y, Dataset)
    assert isinstance(ds.x, Tensor)
    assert isinstance(ds.y.x, Tensor)

    assert "x" in ds._ungrouped_tensors

    ds.create_tensor("/z")
    assert "z" in ds.tensors
    assert "" not in ds.groups
    assert "" not in ds.tensors
    assert isinstance(ds.z, Tensor)

    assert list(ds.groups) == ["y"]
    assert set(ds.tensors) == set(["x", "z", "y/x"])
    assert list(ds.y.tensors) == ["x"]
    z = ds.y.create_group("z")
    assert "z" in ds.y.groups

    c = z.create_tensor("a/b/c")
    d = z.a.b.create_group("d")

    c.append(np.zeros((3, 2)))

    e = ds.create_tensor("/y/z//a/b////d/e/")
    e.append(np.ones((4, 3)))

    ds = local_ds_generator()
    c = ds.y.z.a.b.c
    assert ds.y.z.a.b.parent.group_index == ds.y.z.a.group_index
    np.testing.assert_array_equal(c[0].numpy(), np.zeros((3, 2)))
    assert "d" in ds.y.z.a.b.groups
    e = ds.y.z.a.b.d.e
    np.testing.assert_array_equal(e[0].numpy(), np.ones((4, 3)))

    ds.create_group("g")
    ds.g.create_tensor("g")

    with ds:
        ds.create_group("h")
        ds.h.create_group("i")
        ds.h.i.create_tensor("j")
        assert not ds.storage.autoflush
    assert "j" in ds.h.i.tensors
    assert ds.storage.autoflush


def test_tensor_delete(local_ds_generator):
    ds = local_ds_generator()
    ds.create_tensor("x", max_chunk_size=2 * MB)
    ds.x.extend(np.ones((3, 253, 501, 5)))
    ds.delete_tensor("x")
    assert list(ds.storage.keys()) == ["dataset_meta.json"]
    assert ds.tensors == {}

    ds.create_tensor("x/y")
    ds.delete_tensor("x/y")
    ds.create_tensor("x/y")
    ds["x"].delete_tensor("y")
    ds.delete_group("x")
    assert list(ds.storage.keys()) == ["dataset_meta.json"]
    assert ds.tensors == {}

    ds.create_tensor("x/y/z")
    ds.delete_group("x")
    ds.create_tensor("x/y/z")
    ds["x"].delete_group("y")
    ds.create_tensor("x/y/z")
    ds["x/y"].delete_tensor("z")
    ds.delete_group("x")
    assert list(ds.storage.keys()) == ["dataset_meta.json"]
    assert ds.tensors == {}


def test_group_delete_bug(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("abc/first")
        ds.delete_group("abc")

    ds = local_ds_generator()
    assert ds.tensors == {}
    assert ds.groups == {}


def test_tensor_rename(local_ds_generator):
    ds = local_ds_generator()
    ds.create_tensor("x/y/z")
    ds["x/y/z"].append([1, 2, 3])
    ds.rename_tensor("x/y/z", "x/y/y")

    np.testing.assert_array_equal(ds["x/y/y"][0].numpy(), np.array([1, 2, 3]))

    with pytest.raises(TensorDoesNotExistError):
        ds["x/y/z"].numpy()

    ds.create_tensor("x/y/z")
    ds["x/y/z"].append([4, 5, 6])
    np.testing.assert_array_equal(ds["x/y/z"][0].numpy(), np.array([4, 5, 6]))

    with pytest.raises(RenameError):
        ds.rename_tensor("x/y/y", "x/a")

    with pytest.raises(RenameError):
        ds["x"].rename_tensor("y/y", "y")

    ds.create_tensor("x/y/a/b")
    with pytest.raises(TensorGroupAlreadyExistsError):
        ds["x/y"].rename_tensor("y", "a")

    ds.create_tensor("abc/xyz")
    with pytest.raises(InvalidTensorNameError):
        ds.rename_tensor("abc/xyz", "abc/append")

    ds.create_tensor("abc/efg")
    with pytest.raises(TensorAlreadyExistsError):
        ds.rename_tensor("abc/xyz", "abc/efg")

    ds["x"].rename_tensor("y/y", "y/b")

    np.testing.assert_array_equal(ds["x/y/b"][0].numpy(), np.array([1, 2, 3]))

    ds = local_ds_generator()
    np.testing.assert_array_equal(ds["x/y/b"][0].numpy(), np.array([1, 2, 3]))

    ds.delete_tensor("x/y/b")


def test_group_rename(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("g1/g2/g3/g4/t1")
        ds.create_group("g1/g2/g6")
        ds.create_tensor("g1/g2/t")
        ds["g1/g2/g3/g4/t1"].append([1, 2, 3])
        ds["g1/g2"].rename_group("g3/g4", "g3/g5")
        np.testing.assert_array_equal(
            ds["g1/g2/g3/g5/t1"].numpy(), np.array([[1, 2, 3]])
        )
        with pytest.raises(TensorGroupDoesNotExistError):
            ds["g1"].rename_group("g2/g4", "g2/g5")
        with pytest.raises(TensorGroupAlreadyExistsError):
            ds["g1"].rename_group("g2/g3", "g2/g6")
        with pytest.raises(TensorAlreadyExistsError):
            ds["g1"].rename_group("g2/g3", "g2/t")
        with pytest.raises(InvalidTensorGroupNameError):
            ds["g1"].rename_group("g2/g3", "g2/append")
        with pytest.raises(RenameError):
            ds["g1"].rename_group("g2/g3", "g/g4")
        ds["g1"].rename_group("g2", "g6")
        np.testing.assert_array_equal(
            ds["g1/g6/g3/g5/t1"].numpy(), np.array([[1, 2, 3]])
        )

    with local_ds_generator() as ds:
        np.testing.assert_array_equal(
            ds["g1/g6/g3/g5/t1"].numpy(), np.array([[1, 2, 3]])
        )


def test_vc_bug(local_ds_generator):
    ds = local_ds_generator()
    ds.create_tensor("abc")
    ds.abc.append(1)
    a = ds.commit("first")
    ds.checkout(a)
    ds.create_tensor("a/b/c/d")
    assert list(ds.tensors) == ["abc", "a/b/c/d"]


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
def test_tobytes(memory_ds, compressed_image_paths, audio_paths):
    ds = memory_ds
    ds.create_tensor("image", sample_compression="jpeg")
    ds.create_tensor("audio", sample_compression="mp3")
    with ds:
        for _ in range(3):
            ds.image.append(hub.read(compressed_image_paths["jpeg"][0]))
            ds.audio.append(hub.read(audio_paths["mp3"]))
    with open(compressed_image_paths["jpeg"][0], "rb") as f:
        image_bytes = f.read()
    with open(audio_paths["mp3"], "rb") as f:
        audio_bytes = f.read()
    for i in range(3):
        assert ds.image[i].tobytes() == image_bytes
        assert ds.audio[i].tobytes() == audio_bytes


def test_tobytes_link(memory_ds):
    with memory_ds as ds:
        ds.create_tensor("images", htype="link[image]")
        ds.images.append(hub.link("https://picsum.photos/id/237/200/300"))
        sample = hub.read("https://picsum.photos/id/237/200/300")
        assert ds.images[0].tobytes() == sample.buffer


def test_tensor_clear(local_ds_generator):
    ds = local_ds_generator()
    a = ds.create_tensor("a")
    a.extend([1, 2, 3, 4])
    a.clear()
    assert len(ds) == 0
    assert len(a) == 0

    image = ds.create_tensor("image", htype="image", sample_compression="png")
    image.extend(np.ones((4, 224, 224, 3), dtype="uint8"))
    image.extend(np.ones((4, 224, 224, 3), dtype="uint8"))
    image.clear()
    assert len(ds) == 0
    assert len(image) == 0
    assert image.htype == "image"
    assert image.meta.sample_compression == "png"
    image.extend(np.ones((4, 224, 224, 3), dtype="uint8"))
    a.append([1, 2, 3])

    ds = local_ds_generator()
    assert len(ds) == 1
    assert len(image) == 4
    assert image.htype == "image"
    assert image.meta.sample_compression == "png"


def test_tensor_clear_seq(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("abc", htype="sequence")
        ds.abc.extend([[1, 2, 3, 4]])
        ds.abc.extend([[1, 2, 3, 4, 5]])
        ds.abc.clear()
        assert ds.abc.shape == (0, 0)


def test_no_view(memory_ds):
    memory_ds.create_tensor("a")
    memory_ds.a.extend([0, 1, 2, 3])
    memory_ds.create_tensor("b")
    memory_ds.b.extend([4, 5, 6])
    memory_ds.create_tensor("c/d")
    memory_ds["c/d"].append([7, 8, 9])

    with pytest.raises(InvalidOperationError):
        memory_ds[:2].create_tensor("c")

    with pytest.raises(InvalidOperationError):
        memory_ds[:3].create_tensor_like("c", memory_ds.a)

    with pytest.raises(InvalidOperationError):
        memory_ds[:2].delete_tensor("a")

    with pytest.raises(InvalidOperationError):
        memory_ds[:2].delete_tensor("c/d")

    with pytest.raises(InvalidOperationError):
        memory_ds[:2].delete_group("c")

    with pytest.raises(InvalidOperationError):
        memory_ds[:2].delete()

    with pytest.raises(InvalidOperationError):
        memory_ds.a[:2].append(0)

    with pytest.raises(InvalidOperationError):
        memory_ds.b[:3].extend([3, 4])

    with pytest.raises(InvalidOperationError):
        memory_ds[1:3].read_only = False

    with pytest.raises(InvalidOperationError):
        memory_ds[0].read_only = True

    memory_ds.read_only = True


@pytest.mark.parametrize(
    "x_args", [{}, {"sample_compression": "lz4"}, {"chunk_compression": "lz4"}]
)
@pytest.mark.parametrize(
    "y_args", [{}, {"sample_compression": "lz4"}, {"chunk_compression": "lz4"}]
)
@pytest.mark.parametrize("x_size", [5, (32 * 1000)])
@pytest.mark.parametrize("htype", ["generic", "sequence"])
def test_ds_append(memory_ds, x_args, y_args, x_size, htype):
    ds = memory_ds
    ds.create_tensor("x", **x_args, max_chunk_size=2**20, htype=htype)
    ds.create_tensor("y", dtype="uint8", htype=htype, **y_args)
    with pytest.raises(TensorDtypeMismatchError):
        ds.append({"x": np.ones(2), "y": np.zeros(1)})
    ds.append({"x": np.ones(2), "y": [1, 2, 3]})
    ds.create_tensor("z", htype=htype)
    with pytest.raises(KeyError):
        ds.append({"x": np.ones(2), "y": [4, 5, 6, 7]})
    ds.append({"x": np.ones(3), "y": [8, 9, 10]}, skip_ok=True)
    ds.append({"x": np.ones(4), "y": [2, 3, 4]}, skip_ok=True)
    with pytest.raises(ValueError):
        ds.append({"x": np.ones(2), "y": [4, 5], "z": np.ones(4)})
    with pytest.raises(TensorDtypeMismatchError):
        ds.append({"x": np.ones(x_size), "y": np.zeros(2)}, skip_ok=True)
    assert len(ds.x) == 3
    assert len(ds.y) == 3
    assert len(ds.z) == 0
    assert ds.x.chunk_engine.commit_diff.num_samples_added == 3
    assert ds.y.chunk_engine.commit_diff.num_samples_added == 3
    assert ds.z.chunk_engine.commit_diff.num_samples_added == 0
    assert len(ds) == 0
    for _ in range(3):
        ds.append({"z": np.zeros(2)}, skip_ok=True)
    assert len(ds.z) == 3
    ds.append({"x": np.ones(3), "y": [1, 2, 3]}, append_empty=True)
    assert len(ds.x) == 4
    assert len(ds.y) == 4
    assert len(ds.z) == 4
    assert ds.x.chunk_engine.commit_diff.num_samples_added == 4
    assert ds.y.chunk_engine.commit_diff.num_samples_added == 4
    assert ds.z.chunk_engine.commit_diff.num_samples_added == 4
    assert len(ds) == 4


def test_ds_append_with_ds_view():
    ds1 = hub.dataset("mem://x")
    ds2 = hub.dataset("mem://y")
    ds1.create_tensor("x")
    ds2.create_tensor("x")
    ds1.create_tensor("y")
    ds2.create_tensor("y")
    ds1.append({"x": [0, 1], "y": [1, 2]})
    ds2.append(ds1[0])
    np.testing.assert_array_equal(ds1.x, np.array([[0, 1]]))
    np.testing.assert_array_equal(ds1.x, ds2.x)
    np.testing.assert_array_equal(ds1.y, np.array([[1, 2]]))
    np.testing.assert_array_equal(ds1.y, ds2.y)


def test_ds_extend():
    ds1 = hub.dataset("mem://x")
    ds2 = hub.dataset("mem://y")
    ds1.create_tensor("x")
    ds2.create_tensor("x")
    ds1.create_tensor("y")
    ds2.create_tensor("y")
    ds1.extend({"x": [0, 1, 2, 3], "y": [4, 5, 6, 7]})
    ds2.extend(ds1)
    np.testing.assert_array_equal(ds1.x, np.arange(4).reshape(-1, 1))
    np.testing.assert_array_equal(ds1.x, ds2.x)
    np.testing.assert_array_equal(ds1.y, np.arange(4, 8).reshape(-1, 1))
    np.testing.assert_array_equal(ds1.y, ds2.y)


@pytest.mark.parametrize(
    "src_args", [{}, {"sample_compression": "png"}, {"chunk_compression": "png"}]
)
@pytest.mark.parametrize(
    "dest_args", [{}, {"sample_compression": "png"}, {"chunk_compression": "png"}]
)
@pytest.mark.parametrize("size", [(30, 40, 3), (1261, 759, 3)])
def test_append_with_tensor(src_args, dest_args, size):
    ds1 = hub.dataset("mem://ds1")
    ds2 = hub.dataset("mem://ds2")
    ds1.create_tensor("x", **src_args, max_chunk_size=2 * MB, tiling_threshold=2 * MB)
    x = np.random.randint(0, 256, size, dtype=np.uint8)
    ds1.x.append(x)
    ds2.create_tensor("y", max_chunk_size=3 * MB, tiling_threshold=2 * MB, **dest_args)
    ds2.y.append(ds1.x[0])
    np.testing.assert_array_equal(ds1.x.numpy(), ds2.y.numpy())

    with pytest.raises(SampleAppendingError):
        ds1.append(np.zeros((416, 416, 3)))

    with pytest.raises(SampleAppendingError):
        ds1.append(set())

    with pytest.raises(SampleAppendingError):
        ds1.append([1, 2, 3])


def test_extend_with_tensor():
    ds1 = hub.dataset("mem://ds1")
    ds2 = hub.dataset("mem://ds2")
    with ds1:
        ds1.create_tensor("x")
        ds1.x.extend([1, 2, 3, 4])
    with ds2:
        ds2.create_tensor("x")
        ds2.x.extend(ds1.x)
    np.testing.assert_array_equal(ds1.x, ds2.x)


def test_empty_extend(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("x")
        ds.x.append(1)
        ds.create_tensor("y")
        ds.y.extend(np.zeros((len(ds), 3)))
    assert len(ds) == 0


def test_extend_with_progressbar():
    ds1 = hub.dataset("mem://ds1")
    with ds1:
        ds1.create_tensor("x")
        ds1.x.extend([1, 2, 3, 4], progressbar=True)
    np.testing.assert_array_equal(ds1.x, np.array([[1], [2], [3], [4]]))


def test_auto_htype(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("a")
        ds.create_tensor("b")
        ds.create_tensor("c")
        ds.create_tensor("d")
        ds.create_tensor("e")
        ds.create_tensor("f")
        ds.a.append("hello")
        ds.b.append({"a": [1, 2]})
        ds.c.append([1, 2, 3])
        ds.d.append(np.array([{"x": ["a", 1, 2.5]}]))
        ds.e.append(["a", 1, {"x": "y"}, "b"])
        ds.f.append(ds.e[0])
    assert ds.a.htype == "text"
    assert ds.b.htype == "json"
    assert ds.c.htype == "generic"
    assert ds.d.htype == "json"
    assert ds.e.htype == "json"
    assert ds.f.htype == "json"


def test_sample_shape(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("w")
        ds.create_tensor("x")
        ds.create_tensor("y")
        ds.create_tensor("z")
        ds.w.extend(np.zeros((5, 4, 3, 2)))
        ds.x.extend(np.ones((5, 4000, 5000)))
        ds.y.extend([np.zeros((2, 3)), np.ones((3, 2))])
        ds.z.extend([np.ones((5, 4000, 3000)), np.ones((5, 3000, 4000))])
    assert ds.w[0].shape == (4, 3, 2)
    assert ds.x[0].shape == (4000, 5000)
    assert ds.y[0].shape == (2, 3)
    assert ds.y[1].shape == (3, 2)
    assert ds.z[0].shape == (5, 4000, 3000)
    assert ds.z[1].shape == (5, 3000, 4000)
    assert ds.w[0][0, :2].shape == (2, 2)
    assert ds.z[1][:2, 10:].shape == (2, 2990, 4000)


@enabled_remote_storages
def test_hub_remote_read_images(storage, memory_ds, color_image_paths, gdrive_creds):
    image_path = color_image_paths["jpeg"]
    with open(image_path, "rb") as f:
        byts = f.read()

    memory_ds.create_tensor("images", htype="image", sample_compression="jpg")

    image = hub.read("https://picsum.photos/200/300")
    memory_ds.images.append(image)
    assert memory_ds.images[0].shape == (300, 200, 3)

    storage["sample/samplejpg.jpg"] = byts
    image = hub.read(
        f"{storage.root}/sample/samplejpg.jpg",
        creds=gdrive_creds if storage.root.startswith("gdrive://") else None,
    )
    memory_ds.images.append(image)
    assert memory_ds.images[1].shape == (323, 480, 3)

    storage["samplejpg.jpg"] = byts
    image = hub.read(
        f"{storage.root}/samplejpg.jpg",
        creds=gdrive_creds if storage.root.startswith("gdrive://") else None,
    )
    memory_ds.images.append(image)
    assert memory_ds.images[2].shape == (323, 480, 3)


def test_hub_remote_read_gdrive_root(request, memory_ds, gdrive_creds):
    if not is_opt_true(request, GDRIVE_OPT):
        pytest.skip()
    memory_ds.create_tensor("images", htype="image", sample_compression="jpg")
    memory_ds.images.append(hub.read("gdrive://cat.jpeg", creds=gdrive_creds))
    assert memory_ds.images[0].shape == (900, 900, 3)


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
@enabled_remote_storages
def test_hub_remote_read_videos(storage, memory_ds):
    memory_ds.create_tensor("videos", htype="video", sample_compression="mp4")

    video = hub.read(
        "http://commondatastorage.googleapis.com/gtv-videos-bucket/sample/ForBiggerJoyrides.mp4",
    )
    memory_ds.videos.append(video)
    assert memory_ds.videos[0].shape == (361, 720, 1280, 3)

    if isinstance(storage, GCSProvider):
        video = hub.read(
            "gcs://gtv-videos-bucket/sample/ForBiggerJoyrides.mp4",
        )
        memory_ds.videos.append(video)
        assert memory_ds.videos[1].shape == (361, 720, 1280, 3)


@pytest.mark.parametrize("aslist", (True, False))
@pytest.mark.parametrize(
    "args", [{}, {"sample_compression": "png"}, {"chunk_compression": "png"}]
)
@pytest.mark.parametrize(
    "idx", [3, slice(None), slice(5, 9), slice(3, 7, 2), [3, 7, 6, 4]]
)
def test_sequence_htype(memory_ds, aslist, args, idx):
    ds = memory_ds
    with ds:
        ds.create_tensor("x", htype="sequence", **args)
        for _ in range(10):
            ds.x.append([np.ones((2, 7, 3), dtype=np.uint8) for _ in range(5)])
    np.testing.assert_array_equal(
        np.array(ds.x[idx].numpy(aslist=aslist)), np.ones((10, 5, 2, 7, 3))[idx]
    )
    assert ds.x.shape == (10, 5, 2, 7, 3)
    ds.checkout("branch", create=True)
    with ds:
        for _ in range(5):
            ds.x.append([np.ones((2, 7, 3), dtype=np.uint8) for _ in range(5)])
    np.testing.assert_array_equal(
        np.array(ds.x[idx].numpy(aslist=aslist)), np.ones((15, 5, 2, 7, 3))[idx]
    )
    assert ds.x.shape == (15, 5, 2, 7, 3)
    ds.checkout("main")
    np.testing.assert_array_equal(
        np.array(ds.x[idx].numpy(aslist=aslist)), np.ones((10, 5, 2, 7, 3))[idx]
    )
    assert ds.x.shape == (10, 5, 2, 7, 3)


@pytest.mark.parametrize("shape", [(13, 17, 3), (1007, 3001, 3)])
def test_sequence_htype_with_hub_read(local_ds, shape, compressed_image_paths):
    ds = local_ds
    imgs = list(map(hub.read, compressed_image_paths["jpeg"][:3]))
    arrs = np.random.randint(0, 256, (5, *shape), dtype=np.uint8)
    with ds:
        ds.create_tensor("x", htype="sequence[image]", sample_compression="png")
        for i in range(5):
            if i % 2:
                ds.x.append(imgs)
            else:
                ds.x.append(arrs)
    for i in range(5):
        if i % 2:
            for j in range(3):
                np.testing.assert_array_equal(ds.x[i][j].numpy(), imgs[j].array)
        else:
            for j in range(5):
                np.testing.assert_array_equal(ds.x[i][j].numpy(), arrs[j])


def test_sequence_shapes(memory_ds):
    with memory_ds as ds:
        ds.create_tensor("abc", htype="sequence")
        ds.abc.extend([[1, 2, 3], [4, 5, 6, 7]])

        assert ds.abc[0].shape == (3, 1)

        assert ds.abc.shape_interval.lower == (2, 3, 1)
        assert ds.abc.shape_interval.upper == (2, 4, 1)

        ds.create_tensor("xyz", htype="sequence")
        ds.xyz.append([[[1, 2], [3, 4]], [[1, 2, 3], [4, 5, 6]]])

        assert ds.xyz.shape == (1, 2, 2, None)
        assert ds.xyz[0][0].shape == (2, 2)
        assert ds.xyz[0][1].shape == (2, 3)

        assert ds.xyz.shape_interval.lower == (1, 2, 2, 2)
        assert ds.xyz.shape_interval.upper == (1, 2, 2, 3)

        ds.create_tensor("red", htype="sequence")
        ds.red.extend([[4, 5, 6, 7], [1, 2, 3]])

        assert ds.red[0].shape == (4, 1)

        assert ds.red.shape_interval.lower == (2, 3, 1)
        assert ds.abc.shape_interval.upper == (2, 4, 1)


def test_shape_bug(memory_ds):
    ds = memory_ds
    ds.create_tensor("x")
    ds.x.extend(np.ones((5, 9, 2)))
    assert ds.x[1:4, 3:7].shape == (3, 4, 2)


def test_hidden_tensors(local_ds_generator):
    ds = local_ds_generator()
    with ds:
        ds.create_tensor("x", hidden=True)
        ds.x.append(1)
        assert ds.tensors == {}
        ds.create_tensor("y")
        assert list(ds.tensors.keys()) == ["y"]
        ds.y.extend([1, 2])
        assert len(ds) == 2  # length of hidden tensor is not considered
        ds._hide_tensor("y")
    ds = local_ds_generator()
    assert ds.tensors == {}
    assert len(ds) == 0
    with ds:
        ds.create_tensor("w")
        ds.create_tensor("z")
        ds.append({"w": 2, "z": 3})  # hidden tensors not required

    # Test access
    np.testing.assert_array_equal(ds.x, np.array([[1]]))
    np.testing.assert_array_equal(ds.y, np.array([[1], [2]]))

    assert not ds.w.meta.hidden
    assert not ds.z.meta.hidden
    assert ds.x.meta.hidden
    assert ds.y.meta.hidden


@pytest.mark.parametrize("num_workers", [0, 2])
@pytest.mark.parametrize("progressbar", [True, False])
@pytest.mark.parametrize(
    "index", [slice(None), slice(5, None, None), slice(None, 8, 2), 7]
)
@pytest.mark.parametrize("convert_to_pathlib", [True, False])
def test_dataset_copy(
    memory_ds, local_ds, num_workers, progressbar, index, convert_to_pathlib
):
    memory_ds.path = convert_string_to_pathlib_if_needed(
        memory_ds.path, convert_to_pathlib
    )
    local_ds.path = convert_string_to_pathlib_if_needed(
        local_ds.path, convert_to_pathlib
    )

    ds = memory_ds
    with ds:
        ds.create_tensor("images/image1")
        ds.create_tensor("images/image2")
        ds.create_tensor("label")
        ds.create_tensor("nocopy")
        for _ in range(10):
            ds.images.image1.append(np.random.randint(0, 256, (10, 10, 3)))
            ds.images.image2.append(np.random.randint(0, 256, (10, 10, 3)))
            ds.label.append(np.random.randint(0, 10, (1, 10)))
            ds.nocopy.append([0])

    hub.copy(
        ds[index],
        local_ds.path,
        tensors=["images", "label"],
        overwrite=True,
        num_workers=num_workers,
        progressbar=progressbar,
    )
    local_ds = hub.load(local_ds.path)
    assert set(local_ds.tensors) == set(["images/image1", "images/image2", "label"])
    for t in local_ds.tensors:
        np.testing.assert_array_equal(ds[t][index].numpy(), local_ds[t].numpy())


@pytest.mark.parametrize(
    ("ds_generator", "path", "hub_token"),
    [
        ("local_ds_generator", "local_path", "hub_cloud_dev_token"),
        ("s3_ds_generator", "s3_path", "hub_cloud_dev_token"),
        ("gcs_ds_generator", "gcs_path", "hub_cloud_dev_token"),
        ("hub_cloud_ds_generator", "hub_cloud_path", "hub_cloud_dev_token"),
    ],
    indirect=True,
)
@pytest.mark.parametrize("convert_to_pathlib", [True, False])
def test_hub_exists(ds_generator, path, hub_token, convert_to_pathlib):
    path = convert_string_to_pathlib_if_needed(path, convert_to_pathlib)
    ds = ds_generator()
    assert hub.exists(path, token=hub_token) == True
    assert hub.exists(f"{path}_does_not_exist", token=hub_token) == False


def test_pyav_not_installed(local_ds, video_paths):
    pyav_installed = hub.core.compression._PYAV_INSTALLED
    hub.core.compression._PYAV_INSTALLED = False
    local_ds.create_tensor("videos", htype="video", sample_compression="mp4")
    with pytest.raises(hub.util.exceptions.CorruptedSampleError):
        local_ds.videos.append(hub.read(video_paths["mp4"][0]))
    hub.core.compression._PYAV_INSTALLED = pyav_installed


def test_create_branch_when_locked_out(local_ds):
    local_ds.read_only = True
    local_ds._locked_out = True
    with pytest.raises(ReadOnlyModeError):
        local_ds.create_tensor("x")
    local_ds.checkout("branch", create=True)
    assert local_ds.branch == "branch"
    local_ds.create_tensor("x")


def test_access_method(s3_ds_generator):
    with pytest.raises(DatasetHandlerError):
        hub.dataset("./some_non_existent_path", access_method="download")

    with pytest.raises(DatasetHandlerError):
        hub.dataset("./some_non_existent_path", access_method="local")

    ds = s3_ds_generator()
    with ds:
        ds.create_tensor("x")
        for i in range(10):
            ds.x.append(i)

    ds = s3_ds_generator(access_method="download")
    with pytest.raises(DatasetHandlerError):
        # second time download is not allowed
        s3_ds_generator(access_method="download")
    assert not ds.path.startswith("s3://")
    for i in range(10):
        assert ds.x[i].numpy() == i

    with pytest.raises(ValueError):
        s3_ds_generator(access_method="invalid")

    with pytest.raises(ValueError):
        s3_ds_generator(access_method="download", overwrite=True)

    with pytest.raises(ValueError):
        s3_ds_generator(access_method="local", overwrite=True)

    ds = s3_ds_generator(access_method="local")
    assert not ds.path.startswith("s3://")
    for i in range(10):
        assert ds.x[i].numpy() == i

    ds.delete()


def test_partial_read_then_write(s3_ds_generator):
    ds = s3_ds_generator()
    with ds:
        ds.create_tensor("xyz")
        for i in range(10):
            ds.xyz.append(i * np.ones((1000, 1000)))

    ds = s3_ds_generator()
    np.testing.assert_array_equal(ds.xyz[0].numpy(), 0 * np.ones((1000, 1000)))

    with ds:
        ds.xyz[1] = 20 * np.ones((1000, 1000))


def test_exist_ok(local_ds):
    with local_ds as ds:
        ds.create_tensor("abc")
        with pytest.raises(TensorAlreadyExistsError):
            ds.create_tensor("abc")
        ds.create_tensor("abc", exist_ok=True)
        ds.create_group("grp")
        with pytest.raises(TensorGroupAlreadyExistsError):
            ds.create_group("grp")
        ds.create_group("grp", exist_ok=True)


def verify_label_data(ds):
    text_labels = [
        ["airplane"],
        ["boat"],
        ["airplane"],
        ["car"],
        ["airplane"],
        ["airplane"],
        ["car"],
    ]
    nested_text_labels = [
        [["airplane", "boat", "car"], ["boat", "car", "person"]],
        [["person", "car"], ["airplane", "bus"]],
        [["airplane", "boat"], ["car", "person"]],
    ]
    arr = np.array([0, 1, 0, 2, 0, 0, 2]).reshape((7, 1))
    nested_arr = [
        np.array([[0, 1, 2], [1, 2, 3]]),
        np.array([[3, 2], [0, 4]]),
        np.array([[0, 1], [2, 3]]),
    ]

    random_arr = np.array([[0, 1], [1, 2], [0, 1]])
    random_text_labels = [["l1", "l2"], ["l2", "l3"], ["l1", "l2"]]

    seq_arr = [
        np.array([0, 1, 2]).reshape(3, 1),
        [],
        [],
        np.array([1, 3]).reshape(2, 1),
        np.array([0, 2]).reshape(2, 1),
    ]
    seq_text_labels = [
        [["l1"], ["l2"], ["l3"]],
        [],
        [],
        [["l2"], ["l4"]],
        [["l1"], ["l3"]],
    ]

    # abc
    assert ds.abc.info.class_names == ["airplane", "boat", "car"]
    np_data = ds.abc.numpy()
    data = ds.abc.data()
    assert set(data.keys()) == {"value", "text"}
    np.testing.assert_array_equal(np_data, arr)
    np.testing.assert_array_equal(data["value"], np_data)
    assert data["text"] == text_labels

    # xyz
    assert ds.xyz.info.class_names == []
    np_data = ds.xyz.numpy()
    data = ds.xyz.data()
    assert set(data.keys()) == {"value"}
    np.testing.assert_array_equal(np_data, arr)
    np.testing.assert_array_equal(data["value"], np_data)

    # nested
    assert ds.nested.info.class_names == ["airplane", "boat", "car", "person", "bus"]
    np_data = ds.nested.numpy(aslist=True)
    data = ds.nested.data(aslist=True)
    assert set(data.keys()) == {"value", "text"}
    for i in range(2):
        np.testing.assert_array_equal(np_data[i], nested_arr[i])
        np.testing.assert_array_equal(data["value"][i], np_data[i])
    assert data["text"] == nested_text_labels

    # random
    assert ds.random.info.class_names == ["l1", "l2", "l3", "l4"]
    np_data = ds.random.numpy()
    data = ds.random.data()
    assert set(data.keys()) == {"value", "text"}
    np.testing.assert_array_equal(np_data, random_arr)
    np.testing.assert_array_equal(data["value"], np_data)
    assert data["text"] == random_text_labels

    # seq
    # assert ds.random.info.class_names == ["l1", "l2", "l3", "l4"]
    # np_data = ds.seq.numpy(aslist=True)
    # data = ds.seq.data(aslist=True)
    # assert set(data.keys()) == {"numeric", "text"}
    # for i in range(4):
    #     np.testing.assert_array_equal(np_data[i], seq_arr[i])
    #     np.testing.assert_array_equal(data["numeric"][i], np_data[i])
    # assert data["text"] == seq_text_labels


def test_text_label(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("abc", htype="class_label")
        ds.abc.append("airplane")
        ds.abc.append("boat")
        ds.abc.append("airplane")
        ds.abc.extend(["car", "airplane", 0, 2])

        ds.create_tensor("xyz", htype="class_label")
        ds.xyz.append(0)
        ds.xyz.append(1)
        ds.xyz.append(0)
        ds.xyz.extend([2, 0, 0, 2])

        ds.create_tensor("nested", htype="class_label")
        ds.nested.append([[0, 1, 2], [1, 2, 3]])
        ds.nested.info.class_names = ["airplane", "boat", "car"]
        ds.nested.extend([[["person", 2], ["airplane", "bus"]], [[0, 1], ["car", 3]]])

        temp = hub.constants._ENABLE_RANDOM_ASSIGNMENT
        hub.constants._ENABLE_RANDOM_ASSIGNMENT = True

        ds.create_tensor("random", htype="class_label")
        ds.random[0] = ["l1", "l2"]
        ds.random[1] = ["l2", "l3"]
        ds.random[2] = ["l2", "l4"]
        ds.random[2] = ["l1", "l2"]

        ds.create_tensor("seq", htype="sequence[class_label]")
        ds.seq.append(["l1", "l2", "l3"])
        with pytest.raises(NotImplementedError):
            ds.seq[3] = ["l2", "l4"]
            ds.seq.append(["l3", "l1"])
            ds.seq[4] = ["l1", "l3"]

        hub.constants._ENABLE_RANDOM_ASSIGNMENT = temp

        verify_label_data(ds)

    ds = local_ds_generator()
    verify_label_data(ds)


@pytest.mark.parametrize("num_workers", [0, 2])
def test_text_labels_transform(local_ds_generator, num_workers):
    with local_ds_generator() as ds:
        ds.create_tensor("labels", htype="class_label")
        ds.create_tensor("multiple_labels", htype="class_label")
        ds.create_tensor("seq_labels", htype="sequence[class_label]")

    labels = ["car", "ship", "train"]
    multiple_labels = [["car", "train"], ["ship", "car"], ["train", "ship"]]
    seq_labels = [["ship", "train", "car"], ["ship", "train"], ["car", "train"]]
    data = list(zip(labels, multiple_labels, seq_labels))

    @hub.compute
    def upload(data, ds):
        ds.labels.append(data[0])
        ds.multiple_labels.append(data[1])
        ds.seq_labels.append(data[2])
        return ds

    def convert_to_idx(data, label_idx_map):
        if isinstance(data, str):
            return label_idx_map[data]
        return [convert_to_idx(label, label_idx_map) for label in data]

    upload().eval(data, ds, num_workers=num_workers)

    for tensor in ("labels", "multiple_labels", "seq_labels"):
        class_names = ds[tensor].info.class_names
        label_idx_map = {class_names[i]: i for i in range(len(class_names))}
        if tensor == "labels":
            arr = ds[tensor].numpy()
            assert class_names == ["car", "ship", "train"]
            expected = np.array(convert_to_idx(labels, label_idx_map)).reshape((3, 1))
        elif tensor == "multiple_labels":
            arr = ds[tensor].numpy()
            assert class_names == ["car", "train", "ship"]
            expected = np.array(convert_to_idx(multiple_labels, label_idx_map))
        else:
            arr = ds[tensor].numpy(aslist=True)
            assert class_names == ["ship", "train", "car"]
            expected = convert_to_idx(seq_labels, label_idx_map)
            expected = [np.array(seq).reshape(-1, 1).tolist() for seq in expected]
        np.testing.assert_array_equal(arr, expected)


def test_empty_sample_partial_read(s3_ds):
    with s3_ds as ds:
        ds.create_tensor("xyz")
        ds.xyz.append([1, 2, 3, 4])
        ds.xyz.append(None)
    assert ds.xyz[1].numpy().tolist() == []


def test_htype_config_bug(local_ds):
    with local_ds as ds:
        ds.create_tensor("abc", htype="class_label")
        ds.abc.info.class_names.append("car")
        ds.create_tensor("xyz", htype="class_label")
        assert ds.xyz.info.class_names == []


def test_update_bug(local_ds):
    with local_ds as ds:
        bb = ds.create_tensor("bb", "bbox", dtype="float64")
        arr1 = np.array([[1.0, 2.0, 3.0, 4.0]])
        bb.append(arr1)
        np.testing.assert_array_equal(bb[0].numpy(), arr1)

        arr2 = np.array([[5.0, 6.0, 7.0, 8.0]])
        bb[0] = arr2
        np.testing.assert_array_equal(bb[0].numpy(), arr2)


def test_uneven_view(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("x")
        ds.create_tensor("y")
        ds.x.extend(list(range(10)))
        ds.y.extend(list(range(5)))
        view = ds[list(range(0, 10, 2))]
        np.testing.assert_equal(np.arange(0, 10, 2), view.x.numpy().squeeze())
        with pytest.raises(IndexError):
            np.testing.assert_equal(np.arange(0, 10, 2), view.y.numpy().squeeze())


def test_uneven_iteration(memory_ds):
    with memory_ds as ds:
        ds.create_tensor("x")
        ds.create_tensor("y")
        ds.x.extend(list(range(10)))
        ds.y.extend(list(range(5)))
        ds._enable_padding()
        assert len(ds) == 10
        for i in range(10):
            x, y = ds[i].x.numpy(), ds[i].y.numpy()
            np.testing.assert_equal(x, i)
            target_y = i if i < 5 else []
            np.testing.assert_equal(y, target_y)

        for i, dsv in enumerate(ds):
            x, y = dsv.x.numpy(), dsv.y.numpy()
            np.testing.assert_equal(x, i)
            target_y = i if i < 5 else []
            np.testing.assert_equal(y, target_y)


def token_permission_error_check(
    username,
    password,
    runner,
):
    result = runner.invoke(login, f"-u {username} -p {password}")
    with pytest.raises(TokenPermissionError):
        hub.empty("hub://activeloop-test/sohas-weapons-train")

    with pytest.raises(TokenPermissionError):
        ds = hub.load("hub://activeloop/fake-path")


def invalid_token_exception_check():
    with pytest.raises(InvalidTokenException):
        ds = hub.empty("hub://adilkhan/demo", token="invalid_token")


def user_not_logged_in_exception_check(runner):
    runner.invoke(logout)
    with pytest.raises(UserNotLoggedInException):
        ds = hub.load("hub://activeloop-test/sohas-weapons-train", read_only=True)


def dataset_handler_error_check(runner, username, password):
    result = runner.invoke(login, f"-u {username} -p {password}")
    with pytest.raises(DatasetHandlerError):
        ds = hub.load(f"hub://{username}/wrong-path")


def test_hub_related_permission_exceptions(
    hub_cloud_dev_credentials, hub_cloud_dev_token, hub_dev_token
):
    username, password = hub_cloud_dev_credentials
    runner = CliRunner()

    token_permission_error_check(
        username,
        password,
        runner,
    )
    invalid_token_exception_check()
    user_not_logged_in_exception_check(runner)
    dataset_handler_error_check(runner, username, password)


def test_incompat_dtype_msg(local_ds, capsys):
    local_ds.create_tensor("abc", dtype="uint32")
    with pytest.raises(TensorDtypeMismatchError):
        local_ds.abc.append([0.0])
    captured = capsys.readouterr()
    assert "True" not in captured



from hub.constants import ENCODING_DTYPE
import numpy as np
import hub
import warnings


def version_compare(v1, v2):
    """Returns -1 if v1 is older than v2, 0 if v1 == v2, and +1 if v1 > v2."""

    arr1 = v1.split(".")
    arr2 = v2.split(".")
    n = len(arr1)
    m = len(arr2)

    # converts to integer from string
    arr1 = [int(i) for i in arr1]
    arr2 = [int(i) for i in arr2]

    # compares which list is bigger and fills
    # smaller list with zero (for unequal delimeters)
    if n > m:
        for i in range(m, n):
            arr2.append(0)
    elif m > n:
        for i in range(n, m):
            arr1.append(0)

    # returns 1 if version 1 is bigger and -1 if
    # version 2 is bigger and 0 if equal
    for i in range(len(arr1)):
        if arr1[i] > arr2[i]:
            return 1
        elif arr2[i] > arr1[i]:
            return -1
    return 0


def _check_version(v):
    """Returns True if no fast forwarding is required (False if otherwise)."""

    comparison = version_compare(v, hub.__version__)
    if comparison > 0:
        warnings.warn(
            f"Loading a dataset that was created or updated with a newer version of hub. This could lead to corruption or unexpected errors! Dataset version: {v}, current hub version: {hub.__version__}. It's recommended that you update to a version of hub >= {v}."
        )

    return comparison >= 0


def ffw(func):
    """Decorator for fast forwarding functions. Will handle checking for valid versions and automatically updating the
    fast forwarded version to the latest. Also adds an extra parameter to the decoarted function (version)."""

    def decor(inp, **kwargs):
        v = inp.version
        if not _check_version(v):
            out = func(inp, v, **kwargs)
            inp.version = hub.__version__
            return out

    return decor


@ffw
def ffw_chunk_id_encoder(chunk_id_encoder, version):
    pass


@ffw
def ffw_dataset_meta(dataset_meta, version):
    pass


@ffw
def ffw_tensor_meta(tensor_meta, version):
    versions = ("2.0.2", "2.0.3", "2.0.4", "2.0.5")
    if version in versions and len(tensor_meta.min_shape) == 0:
        # these versions allowed tensors to have a dimensionality of 0
        # newer hub versions require a dimensionality of at least 1
        tensor_meta.min_shape = [1]
        tensor_meta.max_shape = [1]
        tensor_meta.is_dirty = True
    if not hasattr(tensor_meta, "links"):
        tensor_meta.links = {}
    if not hasattr(tensor_meta, "is_link"):
        tensor_meta.is_link = False
    required_meta_keys = tensor_meta._required_meta_keys
    tensor_meta._required_meta_keys = tuple(
        set(required_meta_keys + ("links", "is_link"))
    )


@ffw
def ffw_chunk(chunk, version):
    if version in ("2.0.2", "2.0.3", "2.0.4", "2.0.5"):
        # these versions allowed tensors to have a dimensionality of 0
        # newer hub versions require a dimensionality of at least 1

        shapes = chunk.shapes_encoder
        if shapes.dimensionality == 0:
            a = shapes.array

            # sanity check, this shouldn't be possible
            if len(a) > 1:
                raise ValueError(
                    f"Cannot fast forward an invalid shapes encoder. The length of the encoding was expected to be == 1, but got {len(a)}."
                )

            shapes._encoded = np.array([[1, a[0][0]]], dtype=ENCODING_DTYPE)



# Streams video from a chunk

from typing import Optional, Callable, Tuple
from flask import Flask, request, Response
from hub.core.meta.encode.chunk_id import ChunkIdEncoder
from hub.core.storage import StorageProvider
from hub.core.meta.encode.byte_positions import BytePositionsEncoder
from hub.util.keys import get_chunk_key
from hub.util.remove_cache import get_base_storage
from hub.util.threading import terminate_thread
from hub.util.hash import hash_inputs
from hub.constants import MB
import socketserver
import hub
import threading
import struct
import numpy as np
import re

import logging

log = logging.getLogger("werkzeug")
log.setLevel(logging.ERROR)


_PORT: Optional[int] = None
_SERVER_THREAD: Optional[threading.Thread] = None
_STREAMS = {}
_VIDEO_STREAM_CHUNK_SIZE = 1 * MB


def _get_chunk_key(tensor, index: int) -> str:
    chunk_id = tensor.chunk_engine.chunk_id_encoder[index][0]  # videos are not tiled
    chunk_name = ChunkIdEncoder.name_from_id(chunk_id)
    chunk_commit_id = tensor.chunk_engine.get_chunk_commit(chunk_name)
    chunk_key = get_chunk_key(tensor.key, chunk_name, chunk_commit_id)
    return chunk_key


def _get_stream_key(storage: StorageProvider, chunk_key: str):
    return hash_inputs(storage.root, chunk_key)


def get_video_stream_url(tensor, index: int) -> str:
    chunk_key = _get_chunk_key(tensor, index)
    storage = get_base_storage(tensor.storage)
    stream_key = _get_stream_key(storage, chunk_key)
    if stream_key not in _STREAMS:
        _STREAMS[stream_key] = _VideoStream(storage, chunk_key)
    _start_server()
    index = tensor.chunk_engine.chunk_id_encoder.translate_index_relative_to_chunks(
        index
    )
    return f"http://localhost:{_PORT}/video/{stream_key}/{index}"


class _LazyByteStream:
    MIN_READ_SIZE = 100  # bytes

    def __init__(self, read_f: Callable):
        self.read_f = read_f
        self.buffer = bytearray()

    def read(self, n: int) -> bytes:
        rem = n - len(self.buffer)
        if rem > 0:
            new_bts = self.read_f(max(self.MIN_READ_SIZE, rem))
            ret = self.buffer + new_bts[:rem]
            self.buffer = bytearray(new_bts[rem:])
            return ret
        else:
            ret = self.buffer[:n]
            self.buffer = self.buffer[n:]
            return ret


class _VideoStream:
    __slots__ = [
        "storage",
        "chunk_key",
        "byte_positions_encoder",
        "header_size",
        "chunk_size",
    ]

    def __init__(self, storage: StorageProvider, chunk_key: str):
        self.storage: StorageProvider = storage
        self.chunk_key: str = chunk_key
        self._read_header()

    def _read_header(self):
        offset = {"value": 0}

        def _read(n: int) -> bytes:
            o = offset["value"]
            ret = self.storage.get_bytes(self.chunk_key, o, o + n)
            offset["value"] += n
            return ret

        lazy = _LazyByteStream(_read)
        n_ver = lazy.read(1)[0]
        lazy.read(n_ver)
        r, c = struct.unpack("<ii", lazy.read(8))
        enc_dtype = np.dtype(hub.constants.ENCODING_DTYPE)
        isize = enc_dtype.itemsize
        sh_enc_size = r * c * isize
        lazy.read(sh_enc_size)
        r = int.from_bytes(lazy.read(4), "little")
        bp_enc_size = r * 3 * isize
        self.byte_positions_encoder = BytePositionsEncoder(
            np.frombuffer(lazy.read(bp_enc_size), dtype=enc_dtype).reshape(r, 3).copy()
        )
        self.header_size = 13 + n_ver + sh_enc_size + bp_enc_size
        self.chunk_size = self.storage.get_object_size(self.chunk_key)

    def read(
        self, index: int, start_byte: int, end_byte: int
    ) -> Tuple[bytes, int, int, int]:
        ret_start = start_byte
        sample_start_index, sample_end_index = self.byte_positions_encoder[index]
        offset = self.header_size + sample_start_index
        start_byte += offset
        limit = sample_end_index + self.header_size
        if start_byte >= limit:
            start_byte = offset
            ret_start = 0
        if end_byte is None:
            end_byte = limit
        else:
            end_byte = min(end_byte + offset + 1, limit)

        # Chunking
        chunk_size = end_byte - start_byte
        chunk_size = min(chunk_size, _VIDEO_STREAM_CHUNK_SIZE)
        end_byte = start_byte + chunk_size

        byts = bytes(self.storage.get_bytes(self.chunk_key, start_byte, end_byte))
        return byts, ret_start, len(byts), sample_end_index - sample_start_index


def _get_free_port():
    with socketserver.TCPServer(("localhost", 0), None) as s:
        return s.server_address[1]


def _is_server_running() -> bool:
    return (_SERVER_THREAD is not None) and _SERVER_THREAD.is_alive()


def _run_app():
    try:
        _APP.run(host="0.0.0.0", port=_PORT, threaded=True)
    except Exception:  # thread termination
        pass


def _start_server():
    global _PORT
    global _SERVER_THREAD
    if _is_server_running():
        return
    _PORT = _get_free_port()
    _SERVER_THREAD = threading.Thread(target=_run_app, daemon=True)
    _SERVER_THREAD.start()


def _stop_server():
    global _SERVER_THREAD
    if not _is_server_running():
        return
    request.environ.get("werkzeug.server.shutdown", lambda: None)()
    terminate_thread(_SERVER_THREAD)
    _SERVER_THREAD = None
    _STREAMS.clear()


_APP = Flask("video_stream")


@_APP.after_request
def after_request(response):
    response.headers.add("Accept-Ranges", "bytes")
    return response


@_APP.route("/video/<chunk_id>/<sample_id>")
def stream_video(chunk_id, sample_id):
    range_header = request.headers.get("Range", None)
    start, end = 0, None
    if range_header:
        match = re.search(r"(\d+)-(\d*)", range_header)
        groups = match.groups()

        if groups[0]:
            start = int(groups[0])
        if groups[1]:
            end = int(groups[1])

    chunk, start, length, file_size = _STREAMS[chunk_id].read(
        int(sample_id), start, end
    )
    assert len(chunk) == length

    resp = Response(
        chunk,
        206,
        mimetype="video/mp4",
        content_type="video/mp4",
    )
    resp.headers.add(
        "Connection",
        "keep-alive",
    )
    resp.headers.add("Accept-Ranges", "bytes")
    resp.headers.add(
        "Content-Range",
        "bytes {0}-{1}/{2}".format(start, start + length - 1, file_size),
    )
    return resp



from hub.core.storage.lru_cache import LRUCache
from hub.util.exceptions import InfoError
from hub.core.storage.hub_memory_object import HubMemoryObject
from typing import Any, Dict


class Info(HubMemoryObject):
    def __init__(self):
        self._info = {}
        self._dataset = None

        # the key to info in case of Tensor Info, None in case of Dataset Info
        self._key = None
        self.is_dirty = False

    def __enter__(self):
        from hub.core.tensor import Tensor

        ds = self._dataset
        key = self._key
        if ds is not None:
            ds.storage.check_readonly()
            if not ds.version_state["commit_node"].is_head_node:
                raise InfoError("Cannot modify info from a non-head commit.")
            if key:
                Tensor(key, ds).chunk_engine.commit_diff.modify_info()
            else:
                ds._dataset_diff.modify_info()
            self.is_dirty = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._dataset is not None:
            self._dataset.storage.maybe_flush()

    @property
    def nbytes(self):
        return len(self.tobytes())

    def __len__(self):
        return len(self._info)

    def __getstate__(self) -> Dict[str, Any]:
        return self._info

    def __setstate__(self, state: Dict[str, Any]):
        self._info = state

    def __getattribute__(self, name: str) -> Any:
        """Allows access to info values using the `.` syntax. Example: ``info.description``."""

        if name == "_info":
            return super().__getattribute__(name)
        if name in self._info:
            return self.__getitem__(name)
        return super().__getattribute__(name)

    # implement all the methods of dictionary
    def __getitem__(self, key: str):
        return self._info[key]

    def __str__(self):
        return self._info.__str__()

    def __repr__(self):
        return self._info.__repr__()

    def __setitem__(self, key, value):
        with self:
            self._info[key] = value

    def __delitem__(self, key):
        with self:
            del self._info[key]

    def __contains__(self, key):
        return key in self._info

    def __iter__(self):
        return iter(self._info)

    def __getattr__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            if key == "_info":
                self._info = {}
                return self._info
            return self[key]

    def __setattr__(self, key: str, value):
        if key in {"_info", "_dataset", "_key", "is_dirty"}:
            object.__setattr__(self, key, value)
        else:
            with self:
                self[key] = value

    def get(self, key, default=None):
        return self._info.get(key, default)

    def setdefault(self, key, default=None):
        with self:
            ret = self._info.setdefault(key, default)
        return ret

    def clear(self):
        with self:
            self._info.clear()

    def pop(self, key, default=None):
        with self:
            popped = self._info.pop(key, default)
        return popped

    def popitem(self):
        with self:
            popped = self._info.popitem()
        return popped

    def update(self, *args, **kwargs):
        with self:
            self._info.update(*args, **kwargs)

    def keys(self):
        return self._info.keys()

    def values(self):
        return self._info.values()

    def items(self):
        return self._info.items()

    def replace_with(self, d):
        with self:
            self._info.clear()
            self._info.update(d)

    # the below methods are used by cloudpickle dumps
    def __origin__(self):
        return None

    def __values__(self):
        return None

    def __type__(self):
        return None

    def __union_params__(self):
        return None

    def __tuple_params__(self):
        return None

    def __result__(self):
        return None

    def __args__(self):
        return None


def load_info(path, dataset, key=None) -> Info:
    storage: LRUCache = dataset.storage

    try:
        info = storage.get_hub_object(path, Info)
    except KeyError:
        info = Info()

    info._dataset = dataset
    info._key = key
    storage.register_hub_object(path, info)
    return info
